/*
****************************************************************************
****************************************************************************
**                                                                        **
**                XXXX     X    X         X XXXXXX X                      **
**               X    X   X X   X    X    X X      X   X                  **
**               X       X   X  X   X X   X X      X  X                   **
**               X   XX XXXXXXX X  X   X  X XXXX   X X                    **
**               X    X X     X X X     X X X      XX                     **
**               X    X X     X XX       XX X      X                      **
**                XXXX  X     X X         X XXXXXX XXXXXX                 **
**                                                                        **
**                                                                        **
**                           BIALYSTOK 2007                               **
**                                                                        **
**                        MODULE Frame Util                               **
**                                                                        **
**                     Mikrokontroler ARM: LPC2378                        **
**                                                                        **
**                     Wersja z dnia : 2007.03.22                         **
**                     Napisal : Andrzej Pawluczuk                        **
**                                                                        **
****************************************************************************
****************************************************************************
*/

#include "frameutil.h"
#include "ipservice.h"

USHORT AutoIdentyficationCounter ;

static USHORT CalculateDataChecksum ( ULONG  StartCheckSum ,
                                      void * StartBuffer ,
                                      USHORT BufferSize )
{
  ULONG Sum ;
  USHORT * Buffer ;
  /*-------------------------------------------------------------------------*/
  Sum = StartCheckSum ;
  Buffer = ( USHORT * ) StartBuffer ;
  while ( BufferSize > 1 )
  {
    Sum += * Buffer ++ ;
    BufferSize -= 2 ;
  } /* while */ ;
  if ( BufferSize )
    Sum += * ( UCHAR * ) Buffer ;
  while ( Sum >> 16 )
    Sum = ( Sum & 0xFFFF ) + ( Sum >> 16 ) ;
  return ( ( USHORT ) ( ~ Sum ) ) ;
} /* CalculateDataChecksum */


static USHORT CalculateChecksumWithPseudoHeader ( ULONG  PseudoHeaderCheckSum ,
                                                  void * StartBuffer ,
                                                  USHORT BufferSize )
{
  ULONG Sum ;
  USHORT * Buffer ;
  /*-------------------------------------------------------------------------*/
  Sum = PseudoHeaderCheckSum ;
  Buffer = ( USHORT * ) StartBuffer ;
  while ( BufferSize > 1 )
  {
    Sum += SwapBytes ( * Buffer ++ ) ;
    BufferSize -= 2 ;
  } /* while */ ;
  if ( BufferSize )
  {
      Sum += ( * ( UCHAR * ) Buffer ) << 8 ;
  } /* if */ ;
  while ( Sum >> 16 )
    Sum = ( Sum & 0xFFFF ) + ( Sum >> 16 ) ;
  return ( ( USHORT ) ( ~ Sum ) ) ;
} /* CalculateChecksumWithPseudoHeader */


static ULONG CalcPseudoHeaderCheckSum ( ULONG  SourceIPAddress ,
                                        ULONG  DestinationIPAddress ,
                                        USHORT Protocol ,
                                        USHORT BufferSize )
{
  ULONG CheckSum ;
  /*-------------------------------------------------------------------------*/
  CheckSum = ( SourceIPAddress >> 16 ) + ( SourceIPAddress & 0xFFFF ) ;
  CheckSum += ( DestinationIPAddress >> 16 ) + ( DestinationIPAddress & 0xFFFF ) ;
  CheckSum += ( ULONG ) Protocol + ( ULONG ) BufferSize ;
  return ( CheckSum ) ;
} /* CalcPseudoHeaderCheckSum */


USHORT CalculateChecksum ( void * StartBuffer ,
                           USHORT BufferSize )
{
  /*-------------------------------------------------------------------------*/
  return ( CalculateDataChecksum ( 0 , StartBuffer , BufferSize ) ) ;
} /* CalculateChecksum */


USHORT CalculateUDPChecksum ( ULONG  SourceIPAddress ,
                              ULONG  DestinationIPAddress ,
                              void * StartBuffer ,
                              USHORT BufferSize )
{
  /*-------------------------------------------------------------------------*/
  return ( CalculateChecksumWithPseudoHeader ( CalcPseudoHeaderCheckSum ( SourceIPAddress ,
                                                                          DestinationIPAddress ,
                                                                          Prot_UDP , BufferSize ) ,
                                               StartBuffer , BufferSize ) ) ;
} /* CalculateUDPChecksum */


USHORT CalculateTCPChecksum ( ULONG  SourceIPAddress ,
                              ULONG  DestinationIPAddress ,
                              void * StartBuffer ,
                              USHORT BufferSize )
{
  /*-------------------------------------------------------------------------*/
  return ( CalculateChecksumWithPseudoHeader ( CalcPseudoHeaderCheckSum ( SourceIPAddress ,
                                                                          DestinationIPAddress ,
                                                                          Prot_TCP , BufferSize ) ,
                                               StartBuffer , BufferSize ) ) ;
} /* CalculateTCPChecksum */


void GetByteArrayField ( void *         DestinationAddress ,
                         PtrOSFrameRecT Frame ,
                         USHORT         FieldOffset ,
                         USHORT         FieldSize )
{
  ULONG Loop ;
  UCHAR * SourceData ;
  UCHAR * DestinationData ;
  /*-------------------------------------------------------------------------*/
  SourceData = ( UCHAR * ) Frame -> Data + FieldOffset ;
  DestinationData = ( UCHAR * ) DestinationAddress ;
  for ( Loop = 0 ; Loop < FieldSize ; Loop ++ )
    * DestinationData ++ = * SourceData ++ ;
} /* GetByteArrayField */


UCHAR GetUCHARField ( PtrOSFrameRecT Frame ,
                      USHORT         FieldOffset )
{
  /*-------------------------------------------------------------------------*/
  return ( Frame -> Data [ FieldOffset ] ) ;
} /* GetUCHARField */


USHORT GetUSHORTSwapField ( PtrOSFrameRecT Frame ,
                            USHORT         FieldOffset )
{
  UCHAR Ch1 ;
  UCHAR Ch2 ;
  UCHAR * SourceData ;
  /*-------------------------------------------------------------------------*/
  SourceData = ( UCHAR * ) Frame -> Data + FieldOffset ;
  Ch1 = * SourceData ++ ;
  Ch2 = * SourceData ++ ;
  return ( Ch1 << 8 ) | Ch2 ;
} /* GetUSHORTSwapField */


ULONG GetULONGSwapField ( PtrOSFrameRecT Frame ,
                          USHORT         FieldOffset )
{
  UCHAR Ch1 ;
  UCHAR Ch2 ;
  UCHAR Ch3 ;
  UCHAR Ch4 ;
  UCHAR * SourceData ;
  /*-------------------------------------------------------------------------*/
  SourceData = ( UCHAR * ) Frame -> Data + FieldOffset ;
  Ch1 = * SourceData ++ ;
  Ch2 = * SourceData ++ ;
  Ch3 = * SourceData ++ ;
  Ch4 = * SourceData ++ ;
  return ( Ch1 << 24 ) | ( Ch2 << 16 ) | ( Ch3 << 8 ) | Ch4 ;
} /* GetULONGSwapField */


void AddByteArrayField ( PtrOSFrameRecT Frame ,
                         void *         SourceAddress ,
                         USHORT         FieldSize )
{
  ULONG Loop ;
  UCHAR * SourceData ;
  UCHAR * DestinationData ;
  /*-------------------------------------------------------------------------*/
  DestinationData = ( UCHAR * ) Frame -> Data + Frame -> Length ;
  Frame -> Length += FieldSize ;
  SourceData = ( UCHAR * ) SourceAddress ;
  for ( Loop = 0 ; Loop < FieldSize ; Loop ++ )
    * DestinationData ++ = * SourceData ++ ;
} /* AddByteArrayField */


void AddDuplByteField ( PtrOSFrameRecT Frame ,
                        UCHAR          Data ,
                        USHORT         FieldSize )
{
  ULONG Loop ;
  UCHAR * DestinationData ;
  /*-------------------------------------------------------------------------*/
  DestinationData = ( UCHAR * ) Frame -> Data + Frame -> Length ;
  Frame -> Length += FieldSize ;
  for ( Loop = 0 ; Loop < FieldSize ; Loop ++ )
    * DestinationData ++ = Data ;
} /* AddDuplByteField */


void AddByteField ( PtrOSFrameRecT Frame ,
                    UCHAR          Data )
{
  UCHAR * DestinationData ;
  /*-------------------------------------------------------------------------*/
  DestinationData = ( UCHAR * ) Frame -> Data + Frame -> Length ;
  Frame -> Length ++ ;
  * DestinationData = Data ;
} /* AddByteField */


void AddUSHORTSwapField ( PtrOSFrameRecT Frame ,
                          USHORT         Value )
{
  UCHAR * DestinationData ;
  /*-------------------------------------------------------------------------*/
  DestinationData = ( UCHAR * ) Frame -> Data + Frame -> Length ;
  Frame -> Length += 2 ;
  * DestinationData ++ = ( UCHAR ) ( ( Value >> 8 ) & 0xFF ) ;
  * DestinationData = ( UCHAR ) ( Value & 0xFF ) ;
} /* AddUSHORTSwapField */


void AddULONGSwapField ( PtrOSFrameRecT Frame ,
                         ULONG          Value )
{
  UCHAR * DestinationData ;
  /*-------------------------------------------------------------------------*/
  DestinationData = ( UCHAR * ) Frame -> Data + Frame -> Length ;
  Frame -> Length += 4 ;
  * DestinationData ++ = ( UCHAR ) ( ( Value >> 24 ) & 0xFF ) ;
  * DestinationData ++ = ( UCHAR ) ( ( Value >> 16 ) & 0xFF ) ;
  * DestinationData ++ = ( UCHAR ) ( ( Value >> 8 ) & 0xFF ) ;
  * DestinationData = ( UCHAR ) ( Value & 0xFF ) ;
} /* AddULONGSwapField */


void PutByteArrayField ( PtrOSFrameRecT Frame ,
                         USHORT         FieldOffset ,
                         void *         SourceAddress ,
                         USHORT         FieldSize )
{
  ULONG Loop ;
  UCHAR * SourceData ;
  UCHAR * DestinationData ;
  /*-------------------------------------------------------------------------*/
  DestinationData = ( UCHAR * ) Frame -> Data + FieldOffset ;
  SourceData = ( UCHAR * ) SourceAddress ;
  for ( Loop = 0 ; Loop < FieldSize ; Loop ++ )
    * DestinationData ++ = * SourceData ++ ;
} /* PutByteArrayField */


void PutUSHORTSwapField ( PtrOSFrameRecT Frame ,
                          USHORT         FieldOffset ,
                          USHORT         Value )
{
  UCHAR * DestinationData ;
  /*-------------------------------------------------------------------------*/
  DestinationData = ( UCHAR * ) Frame -> Data + FieldOffset ;
  * DestinationData ++ = ( UCHAR ) ( ( Value >> 8 ) & 0xFF ) ;
  * DestinationData = ( UCHAR ) ( Value & 0xFF ) ;
} /* PutUSHORTSwapField */


void PutUSHORTNoSwapField ( PtrOSFrameRecT Frame ,
                            USHORT         FieldOffset ,
                            USHORT         Value )
{
  UCHAR * DestinationData ;
  /*-------------------------------------------------------------------------*/
  DestinationData = ( UCHAR * ) Frame -> Data + FieldOffset ;
  * DestinationData ++ = ( UCHAR ) ( Value & 0xFF ) ;
  * DestinationData = ( UCHAR ) ( ( Value >> 8 ) & 0xFF ) ;
} /* PutUSHORTNoSwapField */


void StartEthFrame ( PtrOSFrameRecT Frame ,
                     UCHAR *        DestMACAddress ,
                     UCHAR *        SrcMACAddress ,
                     USHORT         LenghtOrType )
{
  ULONG Loop ;
  /*-------------------------------------------------------------------------*/
  Frame -> Length = 0 ;
  Frame -> Index = 0 ;
  if ( DestMACAddress )
    AddByteArrayField ( Frame , DestMACAddress , EthMACAddressSize ) ;
  else
  {
    for ( Loop = 0 ; Loop < EthMACAddressSize ; Loop ++ )
      AddByteField ( Frame , 0x00 ) ;
  } /* if ... else */ ;
  AddByteArrayField ( Frame , SrcMACAddress , EthMACAddressSize ) ;
  AddUSHORTSwapField ( Frame , LenghtOrType ) ;
} /* StartEthFrame */


void CreateIPHeader ( PtrOSFrameRecT Frame ,
                      UCHAR          Protocol ,
                      ULONG          SrcIPAddress ,
                      ULONG          DstIPAddress )
{
  /*-------------------------------------------------------------------------*/
  AddByteField ( Frame , 0x45 ) ;
  AddByteField ( Frame , StandardServiceType ) ;
  AddUSHORTSwapField ( Frame , 0 ) ;        /* Total length -> do uzupelnienia */
  AddUSHORTSwapField ( Frame , AutoIdentyficationCounter ++ ) ;
  AddUSHORTSwapField ( Frame , 0 ) ;        /* FLAGS and fragment offset */
  AddByteField ( Frame , StandardTTL ) ;
  AddByteField ( Frame , Protocol ) ;
  AddUSHORTSwapField ( Frame , 0 ) ;        /* Check sum */
  AddULONGSwapField ( Frame , SrcIPAddress ) ;
  AddULONGSwapField ( Frame , DstIPAddress ) ;
} /* CreateIPHeader */


void SkipByteArrayField ( PtrOSFrameRecT Frame ,
                          USHORT         FieldSize )
{
  /*-------------------------------------------------------------------------*/
  Frame -> Index += FieldSize ;
} /* SkipByteArrayField */


void ReadByteArrayField ( void *         DestinationAddress ,
                          PtrOSFrameRecT Frame ,
                          USHORT         FieldSize )
{
  ULONG Loop ;
  UCHAR * SourceData ;
  UCHAR * DestinationData ;
  /*-------------------------------------------------------------------------*/
  SourceData = ( UCHAR * ) Frame -> Data + Frame -> Index ;
  Frame -> Index += FieldSize ;
  DestinationData = ( UCHAR * ) DestinationAddress ;
  for ( Loop = 0 ; Loop < FieldSize ; Loop ++ )
    * DestinationData ++ = * SourceData ++ ;
} /* ReadByteArrayField */


UCHAR ReadUCHARField ( PtrOSFrameRecT Frame )
{
  UCHAR Ch ;
  /*-------------------------------------------------------------------------*/
  Ch = Frame -> Data [ Frame -> Index ] ;
  Frame -> Index ++ ;
  return ( Ch ) ;
} /* ReadUCHARField */


USHORT ReadUSHORTSwapField ( PtrOSFrameRecT Frame )
{
  UCHAR Ch1 ;
  UCHAR Ch2 ;
  UCHAR * SourceData ;
  /*-------------------------------------------------------------------------*/
  SourceData = ( UCHAR * ) Frame -> Data + Frame -> Index ;
  Frame -> Index += 2 ;
  Ch1 = * SourceData ++ ;
  Ch2 = * SourceData ++ ;
  return ( Ch1 << 8 ) | Ch2 ;
} /* ReadUSHORTSwapField */


ULONG ReadULONGSwapField ( PtrOSFrameRecT Frame )
{
  UCHAR Ch1 ;
  UCHAR Ch2 ;
  UCHAR Ch3 ;
  UCHAR Ch4 ;
  UCHAR * SourceData ;
  /*-------------------------------------------------------------------------*/
  SourceData = ( UCHAR * ) Frame -> Data + Frame -> Index ;
  Frame -> Index += 4 ;
  Ch1 = * SourceData ++ ;
  Ch2 = * SourceData ++ ;
  Ch3 = * SourceData ++ ;
  Ch4 = * SourceData ++ ;
  return ( Ch1 << 24 ) | ( Ch2 << 16 ) | ( Ch3 << 8 ) | Ch4 ;
} /* ReadULONGSwapField */


void GetEthHeader ( EtherPackedHeaderRecT * EtherHeader ,
                    PtrOSFrameRecT          Frame )
{
  /*-------------------------------------------------------------------------*/
  GetByteArrayField ( EtherHeader -> DestinMACAddress , Frame , DestinMACAddressOffset ,
                      EthMACAddressSize ) ;
  GetByteArrayField ( EtherHeader -> SourceMACAddress , Frame , SourceMACAddressOffset ,
                      EthMACAddressSize ) ;
  EtherHeader -> LengthOrType = GetUSHORTSwapField ( Frame , LengthOrTypeOffset ) ;
} /* GetEthHeader */


void GetIPHeader ( IPHeaderRecT * IPHeader ,
                   PtrOSFrameRecT Frame )
{
  UCHAR Temp ;
  /*-------------------------------------------------------------------------*/
  Temp = ReadUCHARField ( Frame ) ;
  IPHeader -> Vers = ( Temp >> 4 ) & 0x0F ;
  IPHeader -> HLen = Temp & 0x0F ;
  IPHeader -> ServiceType = ReadUCHARField ( Frame ) ;
  IPHeader -> TotalLength = ReadUSHORTSwapField ( Frame ) ;
  IPHeader -> Identification = ReadUSHORTSwapField ( Frame ) ;
  IPHeader -> FragmentOffs = ReadUSHORTSwapField ( Frame ) ;
  IPHeader -> Flags = ( UCHAR ) ( ( IPHeader -> FragmentOffs >> 12 ) & 0x0F ) ;
  IPHeader -> FragmentOffs &= 0x0FFF ;
  IPHeader -> TTL = ReadUCHARField ( Frame ) ;
  IPHeader -> Protocol = ReadUCHARField ( Frame ) ;
  IPHeader -> HeaderChecksum = ReadUSHORTSwapField ( Frame ) ;
  IPHeader -> SourceIPAddress = ReadULONGSwapField ( Frame ) ;
  IPHeader -> DestinationIPAddress = ReadULONGSwapField ( Frame ) ;
} /* GetIPHeader */


void SetIPHeaderOrgin ( PtrOSFrameRecT Frame )
{
  /*-------------------------------------------------------------------------*/
  Frame -> Index = EtherPackedHeaderRecTSize ;
} /* SetIPHeaderOrgin */


USHORT SwapBytes ( USHORT Data )
{
  /*-------------------------------------------------------------------------*/
  return ( Data >> 8 ) | ( Data << 8 ) ;
} /* SwapBytes */

