/*
****************************************************************************
****************************************************************************
**                                                                        **
**                XXXX     X    X         X XXXXXX X                      **
**               X    X   X X   X    X    X X      X   X                  **
**               X       X   X  X   X X   X X      X  X                   **
**               X   XX XXXXXXX X  X   X  X XXXX   X X                    **
**               X    X X     X X X     X X X      XX                     **
**               X    X X     X XX       XX X      X                      **
**                XXXX  X     X X         X XXXXXX XXXXXX                 **
**                                                                        **
**                                                                        **
**                           BIALYSTOK 2007                               **
**                                                                        **
**                       MODULE TCP Protocol Service                      **
**                                                                        **
**                     Mikrokontroler ARM: LPC2378                        **
**                                                                        **
**                     Wersja z dnia : 2007.03.22                         **
**                     Napisal : Andrzej Pawluczuk                        **
**                                                                        **
****************************************************************************
****************************************************************************
*/


#include "tcpservice.h"
#include "types.h"
#include "ipservice.h"
#include "frameutil.h"
#include "systime.h"
#include "xramservice.h"

/*
 TCP Header Format  ---> RFC 793

    0                   1                   2                   3   
    0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |          Source Port          |       Destination Port        |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |                        Sequence Number                        |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |                    Acknowledgment Number                      |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |  Data |           |U|A|P|R|S|F|                               |
   | Offset| Reserved  |R|C|S|S|Y|I|            Window             |
   |       |           |G|K|H|T|N|N|                               |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |           Checksum            |         Urgent Pointer        |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |                    Options                    |    Padding    |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
   |                             data                              |
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

Code Bits :
    bit 5 : URG:  Urgent Pointer field significant
    bit 4 : ACK:  Acknowledgment field significant
    bit 3 : PSH:  Push Function
    bit 2 : RST:  Reset the connection
    bit 1 : SYN:  Synchronize sequence numbers
    bit 0 : FIN:  No more data from sender

Options description:
    Currently defined options include:

      Kind     Length    Meaning
      ----     ------    -------
       0         -       End of option list.
       1         -       No-Operation.
       2         4       Maximum Segment Size.

    Specific Option Definitions

      End of Option List

        +--------+
        |00000000|
        +--------+
         Kind=0

      No-Operation

        +--------+
        |00000001|
        +--------+
         Kind=1

      Maximum Segment Size

        +--------+--------+---------+--------+
        |00000010|00000100|   max seg size   |
        +--------+--------+---------+--------+
         Kind=2   Length=4

*/
#define SYN_RectransmitLimit 3

#define StartRoadTime       0x20
#define RetryRoadTime       0x60

#define TCP_Code_FIN        0x01
#define TCP_Code_SYN        0x02
#define TCP_Code_RST        0x04
#define TCP_Code_PSH        0x08
#define TCP_Code_ACK        0x10
#define TCP_Code_URG        0x20

#define TCP_MaxReceivedDataWindow 1024

#define TCPOpionSize_MSS    4
#define TCPOptionCode_MSS   0x0204              // Type 2, Option Length 4 (Max. Segment Size)
#define TCP_MyMSS           1460

#define TCP_SourcePortNoOffset              EtherPackedHeaderRecTSize + IPHeaderSize + 0
#define TCP_DestinationPortNoOffset         EtherPackedHeaderRecTSize + IPHeaderSize + 2
#define TCP_SeqenceNumberOffset             EtherPackedHeaderRecTSize + IPHeaderSize + 4
#define TCP_AcknowledgementNumberOffset     EtherPackedHeaderRecTSize + IPHeaderSize + 8
#define TCP_HdrLengthAndCodeBitsOffset      EtherPackedHeaderRecTSize + IPHeaderSize + 12
#define TCP_WindowOffset                    EtherPackedHeaderRecTSize + IPHeaderSize + 14
#define TCP_TCPCheckSumOffset               EtherPackedHeaderRecTSize + IPHeaderSize + 16
#define TCP_UrgentPointerOffset             EtherPackedHeaderRecTSize + IPHeaderSize + 18

#define TCPHeaderSize                       20


typedef struct { USHORT                 SourcePortNo ;
                 USHORT                 DestinationPortNo ;
                 ULONG                  SeqenceNumber ;
                 ULONG                  AcknowledgementNumber ;
                 USHORT                 HeaderLength ;
                 USHORT                 CodeBits ;
                 USHORT                 Window ;
                 USHORT                 TCPCheckSum ;
                 USHORT                 UrgentPointer ;
                 USHORT                 TCPDataSize ;
               } TCPHeaderRecT ;

#define SocketArraySize 8
static TCPSocketDescriptionRecT SocketArray [ SocketArraySize ] ;
static UCHAR SocketStatus [ SocketArraySize ] ;

#define WindowTableSize  6
static volatile UCHAR WindowTableStatus [ WindowTableSize ] ;
static TCPWindowDescrRecT WindowTable [ WindowTableSize ] ;

static PtrTCPWindowDescrRecT AllocateWindow ( void )
{
  USHORT Loop ;
  /*-------------------------------------------------------------------------*/
  for ( Loop = 0 ; Loop < WindowTableSize ; Loop ++ )
  {
    if ( ! WindowTableStatus [ Loop ] )
    {
      WindowTableStatus [ Loop ] = 1 ;
      return ( & WindowTable [ Loop ] ) ;
    } /* if */ ;
  } /* for */ ;
  return ( AllocateWindowXRAM ( ) ) ;
} /* AllocateWindow */


static void DeallocateWindow ( PtrTCPWindowDescrRecT Window )
{
  USHORT Loop ;
  /*-------------------------------------------------------------------------*/
  for ( Loop = 0 ; Loop < WindowTableSize ; Loop ++ )
  {
    if ( & WindowTable [ Loop ] == Window )
    {
      WindowTableStatus [ Loop ] = 0 ;
      return ;
    } /* if */ ;
  } /* for */ ;
  DeallocateWindowXRAM ( Window ) ;
} /* DeallocateWindow */


static USHORT CheckUniqueSocketNo ( PtrTCPLayerInstanceRecT TCPLayerInstance ,
                                    USHORT                  PortNo )
{
  PtrTCPSocketDescriptionRecT TempSocket ;
  /*-------------------------------------------------------------------------*/
  TempSocket = TCPLayerInstance -> ListenSocketTopEntry ;
  while ( TempSocket )
  {
    if ( TempSocket -> LocalPort == PortNo )
    {
      return ( FALSE ) ;
    } /* if ... */
    else
    {
      TempSocket = TempSocket -> NextSocketLink ;
    } /* if ... else */ ;
  } /* while */ ;
  TempSocket = TCPLayerInstance -> TransmSocketTopEntry ;
  while ( TempSocket )
  {
    if ( TempSocket -> LocalPort == PortNo )
    {
      return ( FALSE ) ;
    } /* if ... */
    else
    {
      TempSocket = TempSocket -> NextSocketLink ;
    } /* if ... else */ ;
  } /* while */ ;
  return ( TRUE ) ;
} /* CheckUniqueSocketNo */


static USHORT AllocTCPPortN ( PtrTCPLayerInstanceRecT TCPLayerInstance )
{
  USHORT PortNo ;
  USHORT Unique ;
  /*-------------------------------------------------------------------------*/
  for ( ; ; )
  {
    PortNo = TCPLayerInstance -> LastLocalPort ++ ;
    if ( TCPLayerInstance -> LastLocalPort > AutoIncrementTCPPortNoUpBound )
      TCPLayerInstance -> LastLocalPort = AutoIncrementTCPPortNoLoBound ;
    Unique = CheckUniqueSocketNo ( TCPLayerInstance , PortNo ) ;
    if ( Unique )
      return ( PortNo ) ;
  } /* for */ ;
} /* AllocTCPPortN */
   

static ULONG GetStartSeqNumber ( PtrTCPLayerInstanceRecT TCPLayerInstance )
{
  /*-------------------------------------------------------------------------*/
  return ( TCPLayerInstance -> InitSequenceNumber + ( TCPLayerInstance -> InitSequenceNumber << 8 ) ) ;
} /* GetStartSeqNumber */


static void StopTimer ( PtrTCPSocketDescriptionRecT Socket )
{
  /*-------------------------------------------------------------------------*/
  Socket -> RetransmData . Counter = 0 ;
  Socket -> RetransmData . DelayTime = 0 ;
} /* StopTimer */


static void StartTimer ( PtrTCPSocketDescriptionRecT Socket ,
                         TCPLastSendType             LastSendOptionType ,
                         USHORT                      Time )
{
  /*-------------------------------------------------------------------------*/
  Socket -> RetransmData . Counter = 0 ;
  Socket -> RetransmData . DelayTime = Time ;
  Socket -> RetransmData . TCPLastSend = LastSendOptionType ;
} /* StopTimer */


static PtrTCPSocketDescriptionRecT AllocateSocket ( PtrTCPLayerInstanceRecT TCPLayerInstance )
{
  USHORT Loop ;
  PtrTCPSocketDescriptionRecT NewSocket ;
  /*-------------------------------------------------------------------------*/
  for ( Loop = 0 ; Loop < SocketArraySize ; Loop ++ )
  {
    if ( ! SocketStatus [ Loop ] )
    {
      SocketStatus [ Loop ] = 1 ;
      NewSocket = & SocketArray [ Loop ] ;
      NewSocket -> NextSocketLink = NIL ;
      NewSocket -> WindowLink = NIL ;
      NewSocket -> TCPLayerInstanceLink = TCPLayerInstance ;
      NewSocket -> TCPStateMachine = CLOSED ;
      NewSocket -> RetransmData . Counter = 0 ;
      NewSocket -> RetransmData . TCPLastSend = None ;
      NewSocket -> RetransmData . DelayTime = 0 ;
      NewSocket -> RemoteAddr = 0 ;
      NewSocket -> RemotePort = 0 ;
      NewSocket -> LocalAddress = 0 ;
      NewSocket -> LocalPort = AllocTCPPortN ( TCPLayerInstance ) ;
      NewSocket -> SND_UNA = 0 ;
      NewSocket -> SND_NXT = 0 ;
      NewSocket -> SND_WND = TCP_MaxReceivedDataWindow  ;
      NewSocket -> SND_UP = 0 ;
      NewSocket -> SND_WL1 = 0 ;
      NewSocket -> SND_WL2 = 0 ;
      NewSocket -> ISS = 0 ;
      NewSocket -> RCV_NXT = 0 ;
      NewSocket -> RCV_WND = TCP_MaxReceivedDataWindow  ;
      NewSocket -> RCV_UP = 0 ;
      NewSocket -> IRS = 0 ;
      NewSocket -> ConnectNotify = NIL ;
      NewSocket -> EstablishNotify = NIL ;
      NewSocket -> ReceiveNotify = NIL ;
      NewSocket -> CloseNotify = NIL ;
      return ( NewSocket ) ;
    } /* if */ ;
  } /* for */ ;
  return ( NIL ) ;
} /* AllocateSocket */


void DeallocateSocket ( PtrTCPSocketDescriptionRecT TCPSocket )
{
  USHORT Loop ;
  PtrTCPWindowDescrRecT TopWindow ;
  PtrTCPWindowDescrRecT DelWindow ;
  /*-------------------------------------------------------------------------*/
  TopWindow = TCPSocket -> WindowLink ;
  TCPSocket -> WindowLink = NIL ;
  while ( TopWindow )
  {
    DelWindow = TopWindow ;
    TopWindow = TopWindow -> FwdLink ;
    DeallocateWindow ( DelWindow ) ;
  } /* while */ ;
  for ( Loop = 0 ; Loop < SocketArraySize ; Loop ++ )
  {
    if ( TCPSocket == ( & SocketArray [ Loop ] ) )
    {
      SocketStatus [ Loop ] = 0 ;
      return ;
    } /* if */ ;
  } /* for */ ;
} /* DeallocateSocket */


void FreeTCBSocket ( PtrTCPSocketDescriptionRecT Socket )
{
  PtrTCPSocketDescriptionRecT TmpSocket ;
  PtrTCPLayerInstanceRecT TCPLayerInstance ;
  /*-------------------------------------------------------------------------*/
  TCPLayerInstance = GetTCPInstance ( Socket ) ;
  if ( TCPLayerInstance -> TransmSocketTopEntry == Socket )
  {
    TCPLayerInstance -> TransmSocketTopEntry = Socket -> NextSocketLink ;
    Socket -> NextSocketLink = NIL ;
  } /* if ... */
  else
  {
    TmpSocket = TCPLayerInstance -> TransmSocketTopEntry ;
    while ( TmpSocket )
    {
      if ( TmpSocket -> NextSocketLink == Socket )
      {
        TmpSocket -> NextSocketLink = Socket -> NextSocketLink ;
        Socket -> NextSocketLink = NIL ;
        break ;
      } /* if ... */
      else
        TmpSocket = TmpSocket -> NextSocketLink ;
    } /* while */ ;
  } /* if ... else */ ;
  DeallocateSocket ( Socket ) ;
} /* FreeTCBSocket */


static void RealDelayedFreeTCBSocket ( void * CallParam )
{
  PtrTCPSocketDescriptionRecT Socket ;
  /*-------------------------------------------------------------------------*/
  Socket = ( PtrTCPSocketDescriptionRecT ) CallParam ;
  FreeTCBSocket ( Socket ) ;
} /* RealDelayedFreeTCBSocket */


static void DelayedFreeTCBSocket ( PtrTCPSocketDescriptionRecT Socket )
{
  /*-------------------------------------------------------------------------*/
  AttachTimer ( 100 , RealDelayedFreeTCBSocket , ( void * ) Socket ) ;
} /* DelayedFreeTCBSocket */


static void AddSocketToListenListe ( PtrTCPSocketDescriptionRecT Socket )
{
  PtrTCPLayerInstanceRecT TCPLayerInstance ;
  /*-------------------------------------------------------------------------*/
  TCPLayerInstance = GetTCPInstance ( Socket ) ;
  Socket -> NextSocketLink = TCPLayerInstance -> ListenSocketTopEntry ;
  TCPLayerInstance -> ListenSocketTopEntry = Socket ;
} /* AddSocketToListenListe */


static void AddSocketToTCBListe ( PtrTCPSocketDescriptionRecT Socket )
{
  PtrTCPLayerInstanceRecT TCPLayerInstance ;
  /*-------------------------------------------------------------------------*/
  TCPLayerInstance = GetTCPInstance ( Socket ) ;
  Socket -> NextSocketLink = TCPLayerInstance -> TransmSocketTopEntry ;
  TCPLayerInstance -> TransmSocketTopEntry = Socket ;
} /* AddSocketToTCBListe */


static ULONG MagicCalculation ( ULONG SeqenceNumber )
{
  /*-------------------------------------------------------------------------*/
  return ( ( ~ SeqenceNumber ) & 0xFFFFF ) ;
} /* MagicCalculation */


static void SendTCPFrame ( PtrTCPSocketDescriptionRecT Socket ,
                           USHORT                      TCPCodeBits )
{
  PtrIPLayerInstanceRecT IPLayerInstance ;
  EMACInstanceRecT * InterfInstance ;
  PtrTCPLayerInstanceRecT TCPLayerInstance ;
  ULONG TargetAddress ;
  PtrOSFrameRecT OutFrame ;
  USHORT CheckSum ;
  void * AllTCPDataBuffer ;
  /*-------------------------------------------------------------------------*/
  TCPLayerInstance = GetTCPInstance ( Socket ) ;
  IPLayerInstance = GetIPInstanceTCP ( TCPLayerInstance ) ;
  InterfInstance = GetEMACInstance ( IPLayerInstance ) ;
  if ( ( Socket -> RemoteAddr & IPLayerInstance -> MyLocalNetMask ) == ( IPLayerInstance -> MyLocalIPAddress & IPLayerInstance -> MyLocalNetMask ) )
  {
    TargetAddress = Socket -> RemoteAddr ;
  } /* if ... */
  else
  {
    TargetAddress = IPLayerInstance -> MyDefaultGWIPAddress ;
    if ( ! TargetAddress )
    {
      return ;
    } /* if */ ;
  } /* if ... else */ ;
  OutFrame = InterfInstance -> InterfaceAllocFrameMemory ( sizeof ( OSFrameRecT ) ) ;
  if ( OutFrame )
  {
    StartEthFrame ( OutFrame , NIL , InterfInstance -> InterfaceMACAddress , FRAME_IP ) ;
    CreateIPHeader ( OutFrame , Prot_TCP , IPLayerInstance -> MyLocalIPAddress , Socket -> RemoteAddr ) ;
    if ( TCPCodeBits & TCP_Code_SYN )
      PutUSHORTSwapField ( OutFrame , IPTotalLengthOffset , IPHeaderSize + TCPHeaderSize + TCPOpionSize_MSS ) ;
    else
      PutUSHORTSwapField ( OutFrame , IPTotalLengthOffset , IPHeaderSize + TCPHeaderSize ) ;
    CheckSum = CalculateChecksum ( & OutFrame -> Data [ IPVersOffset ] , IPHeaderSize ) ;
    PutUSHORTNoSwapField ( OutFrame , IPHeaderChecksumOffset , CheckSum ) ;
    AllTCPDataBuffer = ( void * ) ( OutFrame -> Data + OutFrame -> Length ) ;
    AddUSHORTSwapField ( OutFrame , Socket -> LocalPort ) ;
    AddUSHORTSwapField ( OutFrame , Socket -> RemotePort ) ;
    AddULONGSwapField ( OutFrame , Socket -> SND_NXT ) ;
    AddULONGSwapField ( OutFrame , Socket -> RCV_NXT ) ;
    AddUSHORTSwapField ( OutFrame , 0 ) ;  /* Dlugosc naglowka i Bity Kodu ---> 0 - do uzupelnienia  */
    AddUSHORTSwapField ( OutFrame , Socket -> SND_WND ) ;
    AddUSHORTSwapField ( OutFrame , 0 ) ;  /* Checksum */
    AddUSHORTSwapField ( OutFrame , Socket -> SND_UP ) ;
    if ( TCPCodeBits & TCP_Code_SYN )
    {
      PutUSHORTSwapField ( OutFrame , TCP_HdrLengthAndCodeBitsOffset , 0x6000 | TCPCodeBits ) ;
      AddUSHORTSwapField ( OutFrame , TCPOptionCode_MSS ) ;
      AddUSHORTSwapField ( OutFrame , TCP_MyMSS ) ;
    } /* if ... */
    else
    {
      PutUSHORTSwapField ( OutFrame , TCP_HdrLengthAndCodeBitsOffset , 0x5000 | TCPCodeBits ) ;
    } /* if ... else */ ;
    CheckSum = CalculateTCPChecksum ( IPLayerInstance -> MyLocalIPAddress , Socket -> RemoteAddr ,
                                      AllTCPDataBuffer , 
                                      OutFrame -> Length - EtherPackedHeaderRecTSize - IPHeaderSize ) ;
    PutUSHORTSwapField ( OutFrame , TCP_TCPCheckSumOffset , CheckSum ) ;
    IPLayerInstance -> SendIPPackedService ( IPLayerInstance , OutFrame , TargetAddress ) ;
    InterfInstance -> InterfaceDeallocFrameMemory ( OutFrame ) ;
  } /* if */ ;
} /* SendTCPFrame */


static void SendTCPFrameWithData ( PtrTCPSocketDescriptionRecT Socket )
{
  PtrIPLayerInstanceRecT IPLayerInstance ;
  EMACInstanceRecT * InterfInstance ;
  PtrTCPLayerInstanceRecT TCPLayerInstance ;
  PtrTCPWindowDescrRecT CurrWindow ;
  ULONG TargetAddress ;
  PtrOSFrameRecT OutFrame ;
  USHORT CheckSum ;
  USHORT RdPointer ;
  USHORT Loop ;
  void * AllTCPDataBuffer ;
  /*-------------------------------------------------------------------------*/
  TCPLayerInstance = GetTCPInstance ( Socket ) ;
  IPLayerInstance = GetIPInstanceTCP ( TCPLayerInstance ) ;
  InterfInstance = GetEMACInstance ( IPLayerInstance ) ;
  if ( ( Socket -> RemoteAddr & IPLayerInstance -> MyLocalNetMask ) == ( IPLayerInstance -> MyLocalIPAddress & IPLayerInstance -> MyLocalNetMask ) )
  {
    TargetAddress = Socket -> RemoteAddr ;
  } /* if ... */
  else
  {
    TargetAddress = IPLayerInstance-> MyDefaultGWIPAddress ;
    if ( ! TargetAddress )
    {
      return ;
    } /* if */ ;
  } /* if ... else */ ;
  OutFrame = InterfInstance -> InterfaceAllocFrameMemory ( sizeof ( OSFrameRecT ) ) ;
  if ( OutFrame )
  {
    CurrWindow  = Socket -> WindowLink ;
    StartEthFrame ( OutFrame , NIL , InterfInstance -> InterfaceMACAddress , FRAME_IP ) ;
    CreateIPHeader ( OutFrame , Prot_TCP , IPLayerInstance -> MyLocalIPAddress , Socket -> RemoteAddr ) ;
    PutUSHORTSwapField ( OutFrame , IPTotalLengthOffset , IPHeaderSize + TCPHeaderSize + CurrWindow -> DataLength ) ;
    CheckSum = CalculateChecksum ( & OutFrame -> Data [ IPVersOffset ] , IPHeaderSize ) ;
    PutUSHORTNoSwapField ( OutFrame , IPHeaderChecksumOffset , CheckSum ) ;
    AllTCPDataBuffer = ( void * ) ( OutFrame -> Data + OutFrame -> Length ) ;
    AddUSHORTSwapField ( OutFrame , Socket -> LocalPort ) ;
    AddUSHORTSwapField ( OutFrame , Socket -> RemotePort ) ;
    AddULONGSwapField ( OutFrame , Socket -> SND_NXT ) ;
    AddULONGSwapField ( OutFrame , Socket -> RCV_NXT ) ;
    AddUSHORTSwapField ( OutFrame , 0x5000 | TCP_Code_PSH | TCP_Code_ACK ) ;
    AddUSHORTSwapField ( OutFrame , Socket -> SND_WND ) ;
    AddUSHORTSwapField ( OutFrame , 0 ) ;  /* Checksum */
    AddUSHORTSwapField ( OutFrame , Socket -> SND_UP ) ;
    RdPointer = CurrWindow -> ReadPtr ;
    for ( Loop = 0 ; Loop < CurrWindow -> DataLength ; Loop ++ )
    {
      AddByteField ( OutFrame , CurrWindow -> DataBuffer [ RdPointer ++ ] ) ;
      if ( RdPointer >= MaxWindowSzie )
        RdPointer = 0 ;
    } /* for */ ;
    CheckSum = CalculateTCPChecksum ( IPLayerInstance -> MyLocalIPAddress , Socket -> RemoteAddr ,
                                      AllTCPDataBuffer , 
                                      OutFrame -> Length - EtherPackedHeaderRecTSize - IPHeaderSize ) ;
    PutUSHORTSwapField ( OutFrame , TCP_TCPCheckSumOffset , CheckSum ) ;
    IPLayerInstance -> SendIPPackedService ( IPLayerInstance , OutFrame , TargetAddress ) ;
    InterfInstance -> InterfaceDeallocFrameMemory ( OutFrame ) ;
  } /* if */ ;
} /* SendTCPFrameWithData */


static void RetransmitTCPWindow ( PtrTCPSocketDescriptionRecT Socket )
{
  PtrTCPWindowDescrRecT CurrWindow ;
  /*-------------------------------------------------------------------------*/
  CurrWindow  = Socket -> WindowLink ;
  Socket -> SND_NXT = Socket -> SND_UNA ;
  SendTCPFrameWithData ( Socket ) ;
  StartTimer ( Socket , LastSend_PSH , RetryRoadTime ) ;
  Socket -> SND_NXT += ( ULONG ) CurrWindow -> DataLength ;
} /* RetransmitTCPWindow */


USHORT TCPOpen ( PtrTCPSocketDescriptionRecT  Socket ,
                 ULONG                        RemoteIP ,
                 USHORT                       RemotePort ,
                 TCPEstablishOperationProcT   EstablishCallBackService ,
                 TCPRecvOperationProcT        RecvCallBackService ,
                 TCPCloseOperationProcT       CloseCallBackService )
{
  PtrTCPLayerInstanceRecT TCPLayerInstance ;
  PtrIPLayerInstanceRecT IPLayerInstance ;
  /*-------------------------------------------------------------------------*/
  if ( Socket -> TCPStateMachine != CLOSED )
    return ( TCP_OpenBusy ) ;
  TCPLayerInstance = GetTCPInstance ( Socket ) ;
  IPLayerInstance = GetIPInstanceTCP ( TCPLayerInstance ) ;
  AddSocketToTCBListe ( Socket ) ;
  Socket -> RemoteAddr = RemoteIP ;
  Socket -> RemotePort = RemotePort ;
  Socket -> LocalAddress = IPLayerInstance -> MyLocalIPAddress ;
  Socket -> EstablishNotify = EstablishCallBackService ;
  Socket -> ReceiveNotify = RecvCallBackService ;
  Socket -> CloseNotify = CloseCallBackService ;
  Socket -> ISS = GetStartSeqNumber ( TCPLayerInstance ) ;
  Socket -> SND_UNA = Socket -> ISS ;
  Socket -> SND_NXT = Socket -> ISS ;
  Socket -> SND_UP = 0 ;
  Socket -> SND_WL1 = 0 ;
  Socket -> SND_WL2 = 0 ;
  Socket -> RCV_NXT = 0 ;
  Socket -> RCV_UP = 0 ;
  Socket -> IRS = 0 ;
  SendTCPFrame ( Socket , TCP_Code_SYN ) ;
  StartTimer ( Socket , LastSend_SYN , StartRoadTime ) ;
  Socket -> TCPStateMachine = SYN_SENT ;
  return ( TCP_OpenOK ) ;
} /* TCPOpen */


USHORT TCPClose ( PtrTCPSocketDescriptionRecT Socket )
{
  /*-------------------------------------------------------------------------*/
  if ( Socket -> TCPStateMachine != ESTABLISHED )
    return ( TCP_CloseNotConn ) ;
  SendTCPFrame ( Socket , TCP_Code_FIN | TCP_Code_ACK ) ;
  Socket -> SND_NXT ++ ;
  StartTimer ( Socket , LastSend_FIN , StartRoadTime ) ;
  Socket -> TCPStateMachine = FIN_WAIT_1 ;
  return ( TCP_CloseOK ) ;
} /* TCPClose */


static void SendCurrentSegment ( PtrTCPSocketDescriptionRecT Socket )
{
  PtrTCPWindowDescrRecT CurrWindow ;
  /*-------------------------------------------------------------------------*/
  CurrWindow  = Socket -> WindowLink ;
  SendTCPFrameWithData ( Socket ) ;
  StartTimer ( Socket , LastSend_PSH , StartRoadTime ) ;
  Socket -> SND_NXT += ( ULONG ) CurrWindow -> DataLength ;
} /* SendCurrentSegment */


USHORT TCPSendData ( PtrTCPSocketDescriptionRecT Socket ,
                     void *                      DataBufferPtr ,
                     USHORT                      DataSize )
{
  PtrTCPWindowDescrRecT NewWindow ;
  PtrTCPWindowDescrRecT CurrWindow ;
  USHORT Loop ;
  UCHAR * InputBuffer ;
  /*-------------------------------------------------------------------------*/
  if ( Socket -> TCPStateMachine != ESTABLISHED )
  {
    return ( TCPSend_NoConnect ) ;
  } /* if */ ;
  if ( DataSize > MaxWindowSzie )
  {
    return ( TCPSend_Overflow ) ;
  } /* if */ ;
  InputBuffer = ( UCHAR * ) DataBufferPtr ;
  if ( ! Socket -> WindowLink )
  {
    NewWindow  = AllocateWindow ( ) ;
    if ( ! NewWindow )
    {
      return ( TCPSend_Overflow ) ;
    } /* if */ ;
    NewWindow -> WritePtr = 0 ;
    NewWindow -> ReadPtr = 0 ;
    NewWindow -> DataLength = 0 ;
    NewWindow -> FwdLink = NIL ;
    Socket -> WindowLink = NewWindow ;
    for ( Loop = 0 ; Loop < DataSize ; Loop ++ )
    {
      NewWindow -> DataBuffer [ NewWindow -> WritePtr ++ ] = * InputBuffer ++ ;
      NewWindow -> DataLength ++ ;
    } /* for */ ;
    SendCurrentSegment ( Socket ) ;
    return ( TCPSend_OK ) ;
  } /* if ... */
  else
  {
    CurrWindow = Socket -> WindowLink ;
    NewWindow = CurrWindow -> FwdLink ;
    if ( ! NewWindow )
    {
      NewWindow  = AllocateWindow ( ) ;
      if ( ! NewWindow )
      {
        return ( TCPSend_Overflow ) ;
      } /* if */ ;
      NewWindow -> WritePtr = 0 ;
      NewWindow -> ReadPtr = 0 ;
      NewWindow -> DataLength = 0 ;
      NewWindow -> FwdLink = NIL ;
      CurrWindow -> FwdLink = NewWindow ;
    } /* if */ ;
    if ( ( NewWindow -> DataLength + DataSize ) > MaxWindowSzie )
    {
      return ( TCPSend_Overflow ) ;
    } /* if */ ;
    for ( Loop = 0 ; Loop < DataSize ; Loop ++ )
    {
      NewWindow -> DataBuffer [ NewWindow -> WritePtr ++ ] = * InputBuffer ++ ;
      NewWindow -> DataLength ++ ;
    } /* for */ ;
    return ( TCPSend_OK ) ;
  } /* if ... else */ ;
} /* TCPSendData */


static void EndOfConn ( PtrTCPSocketDescriptionRecT Socket )
{
  /*-------------------------------------------------------------------------*/
  if ( Socket -> CloseNotify )
    Socket -> CloseNotify ( Socket ) ;
  DelayedFreeTCBSocket ( Socket ) ;
//  FreeTCBSocket ( Socket ) ;
} /* EndOfConn */


static void KillSocket ( PtrTCPSocketDescriptionRecT Socket )
{
  /*-------------------------------------------------------------------------*/
  SendTCPFrame ( Socket , TCP_Code_RST ) ;
  EndOfConn ( Socket ) ;
} /* KillSocket */


static void SetStartSocketInfo ( PtrTCPSocketDescriptionRecT TCBSocket ,
                                 IPHeaderRecT *              IPHeader ,
                                 TCPHeaderRecT *             TCPHeader )
{
  /*-------------------------------------------------------------------------*/
  TCBSocket -> RemoteAddr = IPHeader -> SourceIPAddress ;
  TCBSocket -> LocalAddress = IPHeader -> DestinationIPAddress ;
  TCBSocket -> RemotePort = TCPHeader -> SourcePortNo ;
  TCBSocket -> LocalPort = TCPHeader -> DestinationPortNo ;
  TCBSocket -> ISS = MagicCalculation ( TCPHeader -> SeqenceNumber ) ;
  TCBSocket -> SND_UNA = TCBSocket -> ISS ;
  TCBSocket -> SND_NXT = TCBSocket -> ISS ;
  TCBSocket -> SND_WND = TCP_MaxReceivedDataWindow ;
  TCBSocket -> SND_UP = 0 ;     // ??
  TCBSocket -> SND_WL1 = 0 ;    // ??
  TCBSocket -> SND_WL2 = 0 ;    // ??
  TCBSocket -> IRS = TCPHeader -> SeqenceNumber ;
  TCBSocket -> RCV_NXT = TCBSocket -> IRS + 1 ;
  TCBSocket -> RCV_WND = TCPHeader -> Window ;
  TCBSocket -> RCV_UP = TCPHeader -> UrgentPointer ;
} /* SetStartSocketInfo */


static void ReceiveData ( PtrIPLayerInstanceRecT      IPLayerInstance ,
                          PtrTCPSocketDescriptionRecT Socket ,
                          TCPHeaderRecT *             TCPHeader ,
                          PtrOSFrameRecT              Frame )
{
  ULONG Dif ;
  USHORT UserDataSize ;
  UCHAR * TCPBuffer ;
  /*-------------------------------------------------------------------------*/
  TCPBuffer = Frame -> Data + Frame -> Index ;
  if ( Socket -> RCV_NXT == TCPHeader -> SeqenceNumber )
  {
    Socket -> RCV_NXT = Socket -> RCV_NXT + TCPHeader -> TCPDataSize ;
    SendTCPFrame ( Socket , TCP_Code_ACK ) ;
    if ( Socket -> ReceiveNotify )
    {
      Socket -> ReceiveNotify ( Socket , Socket -> RemoteAddr , Socket -> RemotePort ,
                                Socket -> LocalAddress , Socket -> LocalPort ,
                                ( void * ) ( TCPBuffer ) , TCPHeader -> TCPDataSize ) ;
    } /* if */ ;
  } /* if ... */
  else
  {
    Dif = Socket -> RCV_NXT - TCPHeader -> SeqenceNumber ;
    if ( ! ( Dif & 0x80000000 ) )
    {
      if ( ( USHORT ) Dif < TCPHeader -> TCPDataSize )
      {
        UserDataSize = TCPHeader -> TCPDataSize - ( USHORT ) Dif ;
        if ( Socket -> ReceiveNotify )
        {
          Socket -> ReceiveNotify ( Socket , Socket -> RemoteAddr , Socket -> RemotePort ,
                                    Socket -> LocalAddress , Socket -> LocalPort ,
                                    ( void * ) ( TCPBuffer + ( USHORT ) Dif ) , UserDataSize ) ;
        } /* if */ ;
        Socket -> RCV_NXT = TCPHeader -> SeqenceNumber + TCPHeader -> TCPDataSize ;
      } /* if */ ;
      SendTCPFrame ( Socket , TCP_Code_ACK ) ;
    } /* if ... */
    else
    {
      SendTCPFrame ( Socket , TCP_Code_ACK ) ;
    } /* if ... else */ ;
  } /* if ... else */ ;
} /* ReceiveData */


static void AnalyseOptions ( PtrTCPSocketDescriptionRecT Socket ,
                             TCPHeaderRecT *             TCPHeader ,
                             PtrOSFrameRecT              Frame )
{
  /*-------------------------------------------------------------------------*/
  Socket -> RCV_WND = 1460 ;
} /* AnalyseOptions */


static void AcceptLastWindowData ( PtrTCPSocketDescriptionRecT Socket ,
                                   TCPHeaderRecT *             TCPHeader )
{
  PtrTCPWindowDescrRecT CurrWindow ;
  /*-------------------------------------------------------------------------*/
  CurrWindow  = Socket -> WindowLink ;
  if ( CurrWindow )
  {
    if ( TCPHeader -> AcknowledgementNumber == Socket -> SND_NXT )
    {
      StopTimer ( Socket ) ;
      Socket -> SND_UNA = TCPHeader -> AcknowledgementNumber ;
      if ( CurrWindow -> FwdLink )
      {
        Socket -> WindowLink = CurrWindow -> FwdLink ;
        DeallocateWindow ( CurrWindow ) ;
        SendCurrentSegment ( Socket ) ;
      } /* if ... */
      else
      {
        DeallocateWindow ( CurrWindow ) ;
        Socket -> WindowLink = NIL ;
      } /* if ... else */ ;
    } /* if ... */
    else
    {
    } /* if ... else */ ;
  } /* if */ ;
} /* AcceptLastWindowData */


static void TCPReceivedStateMachine ( PtrIPLayerInstanceRecT      IPLayerInstance ,
                                      PtrTCPSocketDescriptionRecT Socket ,
                                      IPHeaderRecT *              IPHeader ,
                                      TCPHeaderRecT *             TCPHeader ,
                                      PtrOSFrameRecT              Frame )
{
//
//                              +---------+ ---------\      active OPEN
//                              |  CLOSED |            \    -----------
//                              +---------+<---------\   \   create TCB
//                                |     ^              \   \  snd SYN
//                   passive OPEN |     |   CLOSE        \   \
//                   ------------ |     | ----------       \   \
//                    create TCB  |     | delete TCB         \   \
//                                V     |                      \   \
//                              +---------+            CLOSE    |    \
//                              |  LISTEN |          ---------- |     |
//                              +---------+          delete TCB |     |
//                   rcv SYN      |     |     SEND              |     |
//                  -----------   |     |    -------            |     V
// +---------+      snd SYN,ACK  /       \   snd SYN          +---------+
// |         |<-----------------           ------------------>|         |
// |   SYN   |                    rcv SYN                     |   SYN   |
// |   RCVD  |<-----------------------------------------------|   SENT  |
// |         |                    snd ACK                     |         |
// |         |------------------           -------------------|         |
// +---------+   rcv ACK of SYN  \       /  rcv SYN,ACK       +---------+
//   |           --------------   |     |   -----------
//   |                  x         |     |     snd ACK
//   |                            V     V
//   |  CLOSE                   +---------+
//   | -------                  |  ESTAB  |
//   | snd FIN                  +---------+
//   |                   CLOSE    |     |    rcv FIN
//   V                  -------   |     |    -------
// +---------+          snd FIN  /       \   snd ACK          +---------+
// |  FIN    |<-----------------           ------------------>|  CLOSE  |
// | WAIT-1  |------------------                              |   WAIT  |
// +---------+          rcv FIN  \                            +---------+
//   | rcv ACK of FIN   -------   |                            CLOSE  |
//   | --------------   snd ACK   |                           ------- |
//   V        x                   V                           snd FIN V
// +---------+                  +---------+                   +---------+
// |FINWAIT-2|                  | CLOSING |                   | LAST-ACK|
// +---------+                  +---------+                   +---------+
//   |                rcv ACK of FIN |                 rcv ACK of FIN |  
//   |  rcv FIN       -------------- |    Timeout=2MSL -------------- |  
//   |  -------              x       V    ------------        x       V
//    \ snd ACK                 +---------+delete TCB         +---------+
//     ------------------------>|TIME WAIT|------------------>| CLOSED  |
//                              +---------+                   +---------+
//
  PtrTCPLayerInstanceRecT TCPLayerInstance ;
  PtrTCPSocketDescriptionRecT TCBSocket ;
  USHORT CallBackResult ;
  /*-------------------------------------------------------------------------*/
  TCPLayerInstance = ( PtrTCPLayerInstanceRecT ) IPLayerInstance -> TCPInstanceEntry ;
  TCPHeader -> CodeBits &= ~ TCP_Code_URG ;
  switch ( Socket -> TCPStateMachine )
  {
    case CLOSED        :
      break ;
    case LISTENING     :
/*
                              +---------+
                              |  LISTEN |
                              +---------+
                   rcv SYN      |     |     SEND
                  -----------   |     |    -------
 +---------+      snd SYN,ACK  /       \   snd SYN          +---------+
 |         |<-----------------           ------------------>|         |
 |   SYN   |                    rcv SYN                     |   SYN   |
 |   RCVD  |<-----------------------------------------------|   SENT  |
 |         |                    snd ACK                     |         |
 |         |                                                |         |
 +---------+                                                +---------+
*/
      if ( TCPHeader -> CodeBits & TCP_Code_RST )
      {
        DelayedFreeTCBSocket ( Socket ) ;
//        FreeTCBSocket ( Socket ) ;
        break ;
      } /* if */ ;
      if ( ! ( TCPHeader -> CodeBits & TCP_Code_SYN ) )
      {
        Socket -> SND_NXT = TCPHeader -> SeqenceNumber ;
        Socket -> RemoteAddr = IPHeader -> SourceIPAddress ;
        Socket -> LocalAddress = IPHeader -> DestinationIPAddress ;
        Socket -> RemotePort = TCPHeader -> SourcePortNo ;
        Socket -> LocalPort = TCPHeader -> DestinationPortNo ;
        SendTCPFrame ( Socket , TCP_Code_RST ) ;
        DelayedFreeTCBSocket ( Socket ) ;
//        FreeTCBSocket ( Socket ) ;
        break ;
      } /* if ... */
      else
      {
        CallBackResult = TRUE ;
        if ( Socket -> ConnectNotify )
        {
          CallBackResult = Socket -> ConnectNotify ( Socket , IPHeader -> SourceIPAddress ,
                                                     TCPHeader -> SourcePortNo ) ;
        } /* if */ ;
        if ( CallBackResult )
        {
          TCBSocket = CreateTCPSocket ( TCPLayerInstance ) ;
          * TCBSocket = * Socket ;
          AnalyseOptions ( TCBSocket , TCPHeader , Frame ) ;
          AddSocketToTCBListe ( TCBSocket ) ;
          TCBSocket -> TCPStateMachine = SYN_RECEIVED ;
          SetStartSocketInfo ( TCBSocket , IPHeader , TCPHeader ) ;
          SendTCPFrame ( TCBSocket , TCP_Code_SYN | TCP_Code_ACK ) ;
          StartTimer ( TCBSocket , LastSend_SYNACC , StartRoadTime ) ;
          FreeTCBSocket ( Socket ) ;
        } /* if ... */
        else
        {
          SetStartSocketInfo ( Socket , IPHeader , TCPHeader ) ;
          SendTCPFrame ( Socket , TCP_Code_RST ) ;
          DelayedFreeTCBSocket ( Socket ) ;
//          FreeTCBSocket ( Socket ) ;
        } /* if ... else */ ;
      } /* if ... else */ ;
      break ;
    case SYN_SENT      :
/*
                                                                ACTIVE OPEN
                                                                    |  
                                                                    V  
 +---------+                                                +---------+
 |         |                                                |         |
 |   SYN   |                    rcv SYN                     |   SYN   |
 |   RCVD  |<-----------------------------------------------|   SENT  |
 |         |                    snd ACK                     |         |
 |         |------------------           -------------------|         |
 +---------+   rcv ACK of SYN  \       /  rcv SYN,ACK       +---------+
               --------------   |     |   -----------                  
                      x         |     |     snd ACK                    
                                V     V                                
                              +---------+                              
                              |  ESTAB  |                              
                              +---------+                              
*/
      if ( TCPHeader -> CodeBits & TCP_Code_RST )
      {
        DelayedFreeTCBSocket ( Socket ) ;
//        FreeTCBSocket ( Socket ) ;
        break ;
      } /* if */ ;
      if ( TCPHeader -> CodeBits & TCP_Code_SYN )
      {
        StopTimer ( Socket ) ;
        Socket -> SND_NXT ++ ;
        Socket -> SND_UNA = Socket -> SND_NXT ;
        Socket -> IRS = TCPHeader -> SeqenceNumber ;
        Socket -> RCV_NXT = Socket -> IRS + 1 ;
        Socket -> RCV_UP = 0 ;
        SendTCPFrame ( Socket , TCP_Code_ACK ) ;
        if ( TCPHeader -> CodeBits & TCP_Code_ACK )
        {
          Socket -> TCPStateMachine = ESTABLISHED ;
          if ( Socket -> EstablishNotify )
            Socket -> EstablishNotify ( Socket ) ;
        } /* if ... */
        else
        {
          Socket -> TCPStateMachine = SYN_RECEIVED ;
        } /* if .... else */ ;
      } /* if */ ;
      break ;
    case SYN_RECEIVED  :
/*
 +---------+
 |         |
 |   SYN   |
 |   RCVD  |
 |         |
 |         |------------------ 
 +---------+   rcv ACK of SYN  \
   |           --------------   |
   |                  x         |
   |                            V
   |  CLOSE                   +---------+                              
   | -------                  |  ESTAB  |                              
   | snd FIN                  +---------+                              
   |
   V
 +---------+
 |  FIN    |
 | WAIT-1  |
 +---------+
*/
      if ( TCPHeader -> CodeBits & TCP_Code_RST )
      {
        DelayedFreeTCBSocket ( Socket ) ;
//        FreeTCBSocket ( Socket ) ;
        break ;
      } /* if */ ;
      if ( TCPHeader -> CodeBits & TCP_Code_SYN )
      {                                      /* czy nie dotarlo wyslane z LISTENING: SYN + ACK ? */
                                             /* Nastapi retransmisja przez zegar */
        break ;
      } /* if */ ;
      if ( TCPHeader -> CodeBits & TCP_Code_ACK )
      {
        if ( TCPHeader -> AcknowledgementNumber == ( Socket -> SND_NXT + 1 ) )
        {                                      /* ACK nie mojego polaczenia */
          StopTimer ( Socket ) ;
          Socket -> SND_NXT ++ ;
          Socket -> SND_UNA = Socket -> SND_NXT ;
          Socket -> TCPStateMachine = ESTABLISHED ;
          if ( Socket -> EstablishNotify )
            Socket -> EstablishNotify ( Socket ) ;
        } /* if ... */
        else
        {
          KillSocket ( Socket ) ;
        } /* if ... else */ ;
      } /* if ... */
      else
      {
        StopTimer ( Socket ) ;
        Socket -> TCPStateMachine = ESTABLISHED ;
        if ( Socket -> EstablishNotify )
          Socket -> EstablishNotify ( Socket ) ;
        TCPReceivedStateMachine ( IPLayerInstance , Socket , IPHeader , TCPHeader , Frame ) ;
      } /* if ... else */ ;
      break ;
    case ESTABLISHED   :
/*
                              +---------+                              
                              |  ESTAB  |                              
                              +---------+                              
                       CLOSE    |     |    rcv FIN                     
                      -------   |     |    -------                     
 +---------+          snd FIN  /       \   snd ACK          +---------+
 |  FIN    |<-----------------           ------------------>|  CLOSE  |
 | WAIT-1  |                                                |   WAIT  |
 +---------+                                                +---------+
*/
      if ( TCPHeader -> CodeBits & TCP_Code_RST )
      {
        EndOfConn ( Socket ) ;
        break ;
      } /* if */ ;
      if ( TCPHeader -> CodeBits & TCP_Code_ACK )
      {
        AcceptLastWindowData ( Socket , TCPHeader ) ;
      } /* if */ ;
      if ( TCPHeader -> TCPDataSize )
      {
        ReceiveData ( IPLayerInstance , Socket , TCPHeader , Frame ) ;
      } /* if */ ;
      if ( TCPHeader -> CodeBits & TCP_Code_FIN )
      {
        StopTimer ( Socket ) ;
//        Socket -> TCPStateMachine = CLOSE_WAIT ;
//        Socket -> RCV_NXT = TCPHeader -> SeqenceNumber + 1 ;
//        SendTCPFrame ( Socket , TCP_Code_ACK ) ;
        Socket -> TCPStateMachine = LAST_ACK ;
        Socket -> RCV_NXT ++ ;
        SendTCPFrame ( Socket , TCP_Code_ACK | TCP_Code_FIN ) ;
        StartTimer ( Socket , LastSend_FIN , StartRoadTime ) ;
      } /* if */ ;
      break ;
    case FIN_WAIT_1    :
/*
 +---------+
 |  FIN    |
 | WAIT-1  |------------------
 +---------+          rcv FIN  \
   | rcv ACK of FIN   -------   |
   | --------------   snd ACK   |
   V        x                   V
 +---------+                  +---------+
 |FINWAIT-2|                  | CLOSING |
 +---------+                  +---------+
*/
      if ( TCPHeader -> CodeBits & TCP_Code_RST )
      {
        EndOfConn ( Socket ) ;
        break ;
      } /* if */ ;
      if ( TCPHeader -> CodeBits & TCP_Code_SYN )
      {
        KillSocket ( Socket ) ;
        break ;
      } /* if */ ;
      StopTimer ( Socket ) ;
      if ( TCPHeader -> CodeBits & TCP_Code_ACK )
      {
        if ( TCPHeader -> CodeBits & TCP_Code_FIN )
        {
          Socket -> RCV_NXT ++ ;
          SendTCPFrame ( Socket , TCP_Code_ACK ) ;
          Socket -> TCPStateMachine = CLOSED ;
          EndOfConn ( Socket ) ;
        } /* if ... */
        else
        {
          Socket -> TCPStateMachine = FIN_WAIT_2 ;
        } /* if ... else */ ;
      } /* if ... */
      else
      {
        if ( TCPHeader -> CodeBits & TCP_Code_FIN )
        {
          Socket -> RCV_NXT ++ ;
          SendTCPFrame ( Socket , TCP_Code_ACK ) ;
          Socket -> TCPStateMachine = CLOSING ;
        } /* if ... */
        else
        {
        } /* if ... else */ ;
      } /* if ... else */ ;
      break ;
    case FIN_WAIT_2    :
/*
 +---------+
 |FINWAIT-2|
 +---------+
   |
   |  rcv FIN
   |  -------
    \ snd ACK                 +---------+
     ------------------------>|TIME WAIT|
                              +---------+
*/
      if ( TCPHeader -> TCPDataSize )
      {
        ReceiveData ( IPLayerInstance , Socket , TCPHeader , Frame ) ;
      } /* if */ ;
      if ( TCPHeader -> CodeBits & TCP_Code_RST )
      {
        EndOfConn ( Socket ) ;
        break ;
      } /* if */ ;
      if ( TCPHeader -> CodeBits & TCP_Code_FIN )
      {
        StopTimer ( Socket ) ;
        Socket -> RCV_NXT ++ ;
        SendTCPFrame ( Socket , TCP_Code_ACK ) ;
        Socket -> TCPStateMachine = CLOSED ;
        if ( Socket -> CloseNotify )
          Socket -> CloseNotify ( Socket ) ;
        DelayedFreeTCBSocket ( Socket ) ;
//        FreeTCBSocket ( Socket ) ;
      } /* if */ ;
      break ;
    case CLOSE_WAIT    :
/*
                                                            +---------+
                                                            |  CLOSE  |
                                                            |   WAIT  |
                                                            +---------+
                                                             CLOSE  |  
                                                            ------- |  
                                                            snd FIN V  
                                                            +---------+
                                                            | LAST-ACK|
                                                            +---------+
*/
      StopTimer ( Socket ) ;
      if ( TCPHeader -> CodeBits & TCP_Code_RST )
      {
        EndOfConn ( Socket ) ;
        break ;
      } /* if */ ;
      SendTCPFrame ( Socket , TCP_Code_FIN ) ;
      StartTimer ( Socket , LastSend_FIN , StartRoadTime ) ;
      Socket -> TCPStateMachine = LAST_ACK ;
      break ;
    case CLOSING       :
/*
                              +---------+
                              | CLOSING |
                              +---------+
                    rcv ACK of FIN |
                    -------------- |
                           x       V
                              +---------+
                              |TIME WAIT|
                              +---------+
*/
      StopTimer ( Socket ) ;
      if ( TCPHeader -> CodeBits & TCP_Code_RST )
      {
        EndOfConn ( Socket ) ;
        break ;
      } /* if */ ;
      if ( TCPHeader -> CodeBits & TCP_Code_ACK )
      {
        Socket -> TCPStateMachine = TIME_WAIT ;
      } /* if */ ;
      break ;
    case LAST_ACK      :
      Socket -> TCPStateMachine = CLOSED ;
      EndOfConn ( Socket ) ;
      break ;
    case TIME_WAIT     :
      Socket -> TCPStateMachine = CLOSED ;
      EndOfConn ( Socket ) ;
      break ;
  } /* switch */ ;
} /* TCPReceivedStateMachine */


static void Reject ( PtrIPLayerInstanceRecT      IPLayerInstance ,
                     PtrTCPSocketDescriptionRecT Socket ,
                     IPHeaderRecT *              IPHeader ,
                     TCPHeaderRecT *             TCPHeader ,
                     PtrOSFrameRecT              Frame )
{
  PtrTCPLayerInstanceRecT TCPLayerInstance ;
  PtrTCPSocketDescriptionRecT TCBSocket ;
  /*-------------------------------------------------------------------------*/
  TCPLayerInstance = ( PtrTCPLayerInstanceRecT ) IPLayerInstance -> TCPInstanceEntry ;
  TCBSocket = CreateTCPSocket ( TCPLayerInstance ) ;
  if ( TCBSocket )
  {
    SetStartSocketInfo ( TCBSocket , IPHeader , TCPHeader ) ;
    TCBSocket -> SND_NXT = TCPHeader -> AcknowledgementNumber ;
    TCBSocket -> RCV_NXT = TCPHeader -> SeqenceNumber ;
    TCBSocket -> ConnectNotify = NIL ;
    TCBSocket -> EstablishNotify = NIL ;
    TCBSocket -> ReceiveNotify = NIL ;
    TCBSocket -> CloseNotify = NIL ;
    SendTCPFrame ( TCBSocket , TCP_Code_RST ) ;
    DelayedFreeTCBSocket ( Socket ) ;
//    FreeTCBSocket ( TCBSocket ) ;
  } /* if */ ;
} /* Reject */


static PtrTCPSocketDescriptionRecT LocateSocket_TCB ( PtrTCPLayerInstanceRecT TCPLayerInstance ,
                                                      ULONG                   SourceIPAddress ,
                                                      USHORT                  SourcePortNo ,
                                                      ULONG                   DestinationIPAddress ,
                                                      USHORT                  DestinationPortNo )
{
  PtrTCPSocketDescriptionRecT TempSocket ;
  /*-------------------------------------------------------------------------*/
  TempSocket = TCPLayerInstance -> TransmSocketTopEntry ;
  while ( TempSocket )
  {
    if ( ( TempSocket -> RemoteAddr == SourceIPAddress ) &&
         ( TempSocket -> RemotePort == SourcePortNo ) &&
         ( TempSocket -> LocalAddress == DestinationIPAddress ) &&
         ( TempSocket -> LocalPort == DestinationPortNo ) )
      break ;
    TempSocket = TempSocket -> NextSocketLink ;
  } /* if */ ;
  return ( TempSocket ) ;
} /* LocateSocket_TCB */ ;


static PtrTCPSocketDescriptionRecT LocateSocket_Listen ( PtrTCPLayerInstanceRecT TCPLayerInstance ,
                                                         ULONG                   DestinationIPAddress ,
                                                         USHORT                  DestinationPortNo )
{
  PtrTCPSocketDescriptionRecT TempSocket ;
  /*-------------------------------------------------------------------------*/
  TempSocket = TCPLayerInstance -> ListenSocketTopEntry ;
  while ( TempSocket )
  {
    if ( ( TempSocket -> LocalAddress == DestinationIPAddress ) &&
         ( TempSocket -> LocalPort == DestinationPortNo ) )
      break ;
    TempSocket = TempSocket -> NextSocketLink ;
  } /* if */ ;
  return ( TempSocket ) ;
} /* LocateSocket_Listen */ ;


static void ProcessTCPPacked ( PtrIPLayerInstanceRecT IPLayerInstance ,
                               IPHeaderRecT *         IPHeader ,
                               PtrOSFrameRecT         Frame )
{
  PtrTCPLayerInstanceRecT TCPLayerInstance ;
  TCPHeaderRecT TCPHeader ;
  PtrTCPSocketDescriptionRecT TCBSocket ;
  USHORT Temp ;
  PtrTCPSocketDescriptionRecT Socket ;
  /*-------------------------------------------------------------------------*/
  TCPLayerInstance = ( PtrTCPLayerInstanceRecT ) IPLayerInstance -> TCPInstanceEntry ;
  TCPHeader . SourcePortNo = ReadUSHORTSwapField ( Frame ) ;
  TCPHeader . DestinationPortNo = ReadUSHORTSwapField ( Frame ) ;
  TCPHeader . SeqenceNumber = ReadULONGSwapField ( Frame ) ;
  TCPHeader . AcknowledgementNumber = ReadULONGSwapField ( Frame ) ;
  Temp = ReadUSHORTSwapField ( Frame ) ;
  TCPHeader . HeaderLength = ( Temp & 0xF000 ) >> 10 ;
  TCPHeader . CodeBits = Temp & 0x3F ;
  TCPHeader . Window = ReadUSHORTSwapField ( Frame ) ;
  TCPHeader . TCPCheckSum = ReadUSHORTSwapField ( Frame ) ;
  TCPHeader . UrgentPointer = ReadUSHORTSwapField ( Frame ) ;
  Temp = IPHeaderSize + TCPHeader . HeaderLength ;
  if ( IPHeader -> TotalLength >= Temp )
  {
    TCPHeader . TCPDataSize = IPHeader -> TotalLength - Temp ;
    if ( TCPHeader . HeaderLength > TCPHeaderSize )
    {
      SkipByteArrayField ( Frame , TCPHeader . HeaderLength - TCPHeaderSize ) ;
    } /* if */ ;
    Socket = LocateSocket_TCB ( TCPLayerInstance , IPHeader -> SourceIPAddress , TCPHeader . SourcePortNo ,
                                IPHeader -> DestinationIPAddress , TCPHeader . DestinationPortNo ) ;
    if ( ! Socket )
    {
      Socket = LocateSocket_Listen ( TCPLayerInstance , IPHeader -> DestinationIPAddress ,
                                     TCPHeader . DestinationPortNo ) ;
      if ( Socket )
      {
        TCBSocket = CreateTCPSocket ( TCPLayerInstance ) ;
        * TCBSocket = * Socket ;
        AddSocketToTCBListe ( TCBSocket ) ;
        Socket = TCBSocket ;
      } /* if */ ;
    } /* if */ ;
    if ( Socket )
    {
      TCPReceivedStateMachine ( IPLayerInstance , Socket , IPHeader , & TCPHeader , Frame ) ;
    } /* if ... */
    else
    {
      Reject ( IPLayerInstance , Socket , IPHeader , & TCPHeader , Frame ) ;
    } /* if ... else */ ;
  } /* if */ ;
} /* ProcessTCPPacked */


PtrTCPSocketDescriptionRecT CreateTCPSocket ( PtrTCPLayerInstanceRecT TCPLayerInstance )

{
  PtrTCPSocketDescriptionRecT NewSocket ;
  /*-------------------------------------------------------------------------*/
  NewSocket = AllocateSocket ( TCPLayerInstance ) ;
  return ( NewSocket ) ;
} /* CreateTCPSocket */


USHORT BindTCPSocket ( PtrTCPSocketDescriptionRecT TCPSocket ,
                       USHORT                      TCPPortNo )
{
  PtrTCPLayerInstanceRecT TCPLayerInstance ;
  USHORT Unique ;
  /*-------------------------------------------------------------------------*/
  TCPLayerInstance = GetTCPInstance ( TCPSocket ) ;
  if ( TCPSocket -> TCPStateMachine != CLOSED )
    return ( TCP_BindPortIsOpened ) ;
  Unique = CheckUniqueSocketNo ( TCPLayerInstance , TCPPortNo ) ;
  if ( ! Unique )
    return ( TCP_BindPortDupl ) ;
  TCPSocket -> LocalPort = TCPPortNo ;
  return ( TCP_BindPortOK ) ;
} /* BindTCPSocket */


USHORT TCPListen ( PtrTCPSocketDescriptionRecT  TCPSocket ,
                   TCPConnetOperationProcT      ConnetCallBackService ,
                   TCPEstablishOperationProcT   EstablishCallBackService ,
                   TCPRecvOperationProcT        RecvCallBackService ,
                   TCPCloseOperationProcT       CloseCallBackService )
{
  PtrIPLayerInstanceRecT IPLayerInstance ;
  PtrTCPLayerInstanceRecT TCPLayerInstance ;
  /*-------------------------------------------------------------------------*/
  TCPLayerInstance = GetTCPInstance ( TCPSocket ) ;
  if ( TCPSocket -> TCPStateMachine != CLOSED )
    return ( TCP_ListenPortBusy ) ;
  IPLayerInstance = GetIPInstanceTCP ( TCPLayerInstance ) ;
  AddSocketToListenListe ( TCPSocket ) ;
  TCPSocket -> TCPStateMachine = LISTENING ;
  TCPSocket -> LocalAddress = IPLayerInstance -> MyLocalIPAddress ;
  TCPSocket -> ConnectNotify = ConnetCallBackService ;
  TCPSocket -> EstablishNotify = EstablishCallBackService ;
  TCPSocket -> ReceiveNotify = RecvCallBackService ;
  TCPSocket -> CloseNotify = CloseCallBackService ;
  return ( TCP_ListenPortOK ) ;
} /* TCPListen */


static void TCPTimeService ( void * TCPInstanceEntry )
{
  PtrTCPLayerInstanceRecT TCPLayerInstance ;
  PtrTCPSocketDescriptionRecT Socket ;
  PtrRetransmRecT RetransmitDescr ;
  /*-------------------------------------------------------------------------*/
  TCPLayerInstance = ( PtrTCPLayerInstanceRecT ) TCPInstanceEntry ;
  Socket = TCPLayerInstance -> TransmSocketTopEntry ;
  while ( Socket )
  {
    TCPLayerInstance -> InitSequenceNumber ++ ;
    RetransmitDescr = & Socket -> RetransmData ;
    if ( RetransmitDescr -> DelayTime )
    {
      RetransmitDescr -> DelayTime -- ;
      if ( ! RetransmitDescr -> DelayTime )
      {
        TCPLayerInstance -> InitSequenceNumber ++ ;
        switch ( RetransmitDescr -> TCPLastSend )
        {
          case None          :
            break ;
          case LastSend_FIN  :
            SendTCPFrame ( Socket , TCP_Code_FIN | TCP_Code_ACK ) ;
            RetransmitDescr -> DelayTime = RetryRoadTime ;
            break ;
          case LastSend_SYN  :
            RetransmitDescr -> Counter ++ ;
            if ( RetransmitDescr -> Counter < SYN_RectransmitLimit )
            {
              SendTCPFrame ( Socket , TCP_Code_SYN ) ;
              RetransmitDescr -> DelayTime = RetryRoadTime ;
            } /* if ... */
            else
            {
              if ( Socket -> CloseNotify )
                Socket -> CloseNotify ( Socket ) ;
              DelayedFreeTCBSocket ( Socket ) ;
//              FreeTCBSocket ( Socket ) ;
              return ;
            } /* if ... else */ ;
            break ;
          case LastSend_SYNACC :
            SendTCPFrame ( Socket , TCP_Code_SYN | TCP_Code_ACK ) ;
            RetransmitDescr -> DelayTime = RetryRoadTime ;
            break ;
          case LastSend_PSH  :
            RetransmitTCPWindow ( Socket ) ;
            break ;
          case LastSend_RST  :
            break ;
        } /* switch */ ;
      } /* if */ ;
    } /* if */ ;
    Socket = Socket -> NextSocketLink ;
  } /* while */ ;
} /* TCPTimeService */


void TCPInitInstance ( PtrTCPLayerInstanceRecT TCPLayerInstance ,
                       PtrIPLayerInstanceRecT  IPLayerInstance )
{
  USHORT Loop ;
  /*-------------------------------------------------------------------------*/
  TCPLayerInstance -> ListenSocketTopEntry = NIL ;
  TCPLayerInstance -> TransmSocketTopEntry = NIL ;
  TCPLayerInstance -> LastLocalPort = AutoIncrementTCPPortNoLoBound ;
  TCPLayerInstance -> InitSequenceNumber = 16347 ;
  TCPLayerInstance -> IPLayerInstance = IPLayerInstance ;
  IPLayerInstance -> TCPProcessService = ProcessTCPPacked ;
  IPLayerInstance -> TCPInstanceEntry = TCPLayerInstance ;
  IPLayerInstance -> TCPTimerPoolService = TCPTimeService ;
  for ( Loop = 0 ; Loop < SocketArraySize ; Loop ++ )
    SocketStatus [ Loop ] = 0 ;
  for ( Loop = 0 ; Loop < WindowTableSize ; Loop ++ )
    WindowTableStatus [ Loop ] = 0 ;
} /* TCPInitInstance */


PtrIPLayerInstanceRecT GetIPInstanceTCP ( PtrTCPLayerInstanceRecT TCPLayerInstance )
{
  /*-------------------------------------------------------------------------*/
  return ( TCPLayerInstance -> IPLayerInstance ) ;
} /* GetIPInstanceTCP */


PtrTCPLayerInstanceRecT GetTCPInstance ( PtrTCPSocketDescriptionRecT TCPSocket )
{
  /*-------------------------------------------------------------------------*/
  return ( TCPSocket -> TCPLayerInstanceLink ) ;
} /* GetTCPInstance */


static TCPLayerInstanceRecT LocalTCPLayerInstance ;
PtrTCPLayerInstanceRecT StandardAddTCPInstance ( PtrIPLayerInstanceRecT IPLayerInstance )
{
  PtrTCPLayerInstanceRecT TCPLayerInstance ;
  /*-------------------------------------------------------------------------*/
  TCPLayerInstance = & LocalTCPLayerInstance ;
  TCPInitInstance ( TCPLayerInstance , IPLayerInstance ) ;
  return ( TCPLayerInstance ) ;
} /* StandardAddTCPInstance */


