/*
****************************************************************************
****************************************************************************
**                                                                        **
**                XXXX     X    X         X XXXXXX X                      **
**               X    X   X X   X    X    X X      X   X                  **
**               X       X   X  X   X X   X X      X  X                   **
**               X   XX XXXXXXX X  X   X  X XXXX   X X                    **
**               X    X X     X X X     X X X      XX                     **
**               X    X X     X XX       XX X      X                      **
**                XXXX  X     X X         X XXXXXX XXXXXX                 **
**                                                                        **
**                           BIALYSTOK 2007                               **
**                                                                        **
**                    MODULE LPC2378 UART0 Service                        **
**                                                                        **
**                     Mikrokontroler ARM: LPC2378                        **
**                                                                        **
**                     Wersja z dnia : 2007.03.25                         **
**                     Napisal : Andrzej Pawluczuk                        **
**                                                                        **
****************************************************************************
****************************************************************************
*/

#include <LPC23xx.h>
#include "uart0.h"
#include "types.h"
#include "system.h"
#include "irqvectors.h"
#include "serialconst.h"
#include "leddisplay.h"

#define DefaultLEDDelay       44

#define UART0InterruptMask ( 1 << UART0_Interrupt )

#define InpSerialBuffSize     64
#define OutSerialBuffSize     128
                              /* 64 */

typedef struct {
                   USHORT     RdDataPointer ;
                   USHORT     WrDataPointer ;
                   UCHAR      Buffer [ InpSerialBuffSize ] ;
                 } InpSerialBufferRecT ;

typedef struct {
                   USHORT     RdDataPointer ;
                   USHORT     WrDataPointer ;
                   USHORT     BusyStatus ;
                   UCHAR      Buffer [ OutSerialBuffSize ] ;
                 } OutSerialBufferRecT ;


static volatile InpSerialBufferRecT UART0InpBuffer ;
static volatile OutSerialBufferRecT UART0OutBuffer ;
UCHAR UART0TxdLEDDelayCounter ;
UCHAR UART0RxDLEDDelayCounter ;

/*
U0RBR - UART0 Receiver Buffer Register (address 0xE000 C000:read operation)
   +-------+-------------------------------------------------------------------------+----------+
   |  bit  |                                                                         |Reset val.|
   +-------+-------------------------------------------------------------------------+----------+
   |  7:0  | The UART0 Receiver Buffer Register contains the oldest received byte in |     ?    |
   |       | the UART0 Rx FIFO.                                                      |          |
   +-------+-------------------------------------------------------------------------+----------+


U0THR - UART0 Transmit Holding Register (0xE000 C000:write operation)
   +-------+-------------------------------------------------------------------------+----------+
   |  bit  |                                                                         |Reset val.|
   +-------+-------------------------------------------------------------------------+----------+
   |  7:0  | THR Writing to the UART0 Transmit Holding Register causes the data      |      ?   |
   |       | to be stored in the UART0 transmit FIFO. The byte will be sent when it  |          |
   |       | reaches the bottom of the FIFO and the transmitter is available.        |          |
   +-------+-------------------------------------------------------------------------+----------+


U0DLL - UART0 Divisor Latch LSB Register (address 0xE000 C000)
   +-------+-------------------------------------------------------------------------+----------+
   |  bit  |                                                                         |Reset val.|
   +-------+-------------------------------------------------------------------------+----------+
   |  7:0  | DLLSB: The UART0 Divisor Latch LSB Register                             |     1    |
   +-------+-------------------------------------------------------------------------+----------+


U0DLM - UART0 Divisor Latch MSB Register (address 0xE000 C004)
   +-------+-------------------------------------------------------------------------+----------+
   |  bit  |                                                                         |Reset val.|
   +-------+-------------------------------------------------------------------------+----------+
   |  7:0  | DLMSB The UART0 Divisor Latch MSB Register                              |     0    |
   +-------+-------------------------------------------------------------------------+----------+


U0IER - UART0 Interrupt Enable Register (address 0xE000 C004)
   +-------+-------------------------------------------------------------------------+----------+
   |  bit  |                                                                         |Reset val.|
   +-------+-------------------------------------------------------------------------+----------+
   |   0   | RBR Interrupt Enable                                                    |     0    |
   |       |  0 -> Disable the RDA interrupts.                                       |          |
   |       |  1 -> Enable the RDA interrupts.                                        |          |
   +-------+-------------------------------------------------------------------------+----------+
   |   1   | THRE Interrupt Enable                                                   |     0    |
   |       |  0 -> Disable the THRE interrupts.                                      |          |
   |       |  1 -> Enable the THRE interrupts.                                       |          |
   +-------+-------------------------------------------------------------------------+----------+
   |   2   | RX Line Status Interrupt Enable                                         |     0    |
   |       |  0 -> Disable the RX line status interrupts.                            |          |
   |       |  1 -> Enable the RX line status interrupts.                             |          |
   +-------+-------------------------------------------------------------------------+----------+
   |  7:3  | Reserved                                                                |     ?    |
   +-------+-------------------------------------------------------------------------+----------+
   |   8   | ABTOIntEn                                                               |     0    |
   |       |  0 -> Disable Auto-baud Time-out Interrupt.                             |          |
   |       |  1 -> Enable Auto-baud Time-out Interrupt.                              |          |
   +-------+-------------------------------------------------------------------------+----------+
   |   9   | ABEOIntEn                                                               |     0    |
   |       |  0 -> Disable End of Auto-baud Interrupt.                               |          |
   |       |  1 -> Enable End of Auto-baud Interrupt.                                |          |
   +-------+-------------------------------------------------------------------------+----------+
   | 31:10 | Reserved                                                                |     ?    |
   +-------+-------------------------------------------------------------------------+----------+


U0IIR - UART0 Interrupt Identification Register (address 0xE000 C008)
   +-------+-------------------------------------------------------------------------+----------+
   |  bit  |                                                                         |Reset val.|
   +-------+-------------------------------------------------------------------------+----------+
   |   0   | IntStatus                                                               |     1    |
   |       |   1 -> No interrupt is pending.                                         |          |
   +-------+-------------------------------------------------------------------------+----------+
   |  3:1  | IntId                                                                   |     0    |
   |       |   011 -> Receive Line Status (RLS)                                      |          |
   |       |   010 - Receive Data Available (RDA).                                   |          |
   |       |   110 - Character Time-out Indicator (CTI).                             |          |
   |       |   001 3 - THRE Interrupt                                                |          |
   +-------+-------------------------------------------------------------------------+----------+
   |  5:4  | Reserved                                                                |     ?    |
   +-------+-------------------------------------------------------------------------+----------+
   |  7:6  | FIFO Enable:                                                            |     0    |
   |       |   These bits are equivalent to U0FCR                                    |          |
   +-------+-------------------------------------------------------------------------+----------+
   |    8  | ABEOInt:                                                                |     0    |
   |       |   End of auto-baud interrupt. True if auto-baud has finished            |          |
   |       |   successfully and interrupt is enabled.                                |          |
   +-------+-------------------------------------------------------------------------+----------+
   |    9  | ABTOInt:                                                                |     0    |
   |       |   Auto-baud time-out interrupt. True if auto-baud has timed             |          |
   |       |   out and interrupt is enabled.                                         |          |
   +-------+-------------------------------------------------------------------------+----------+
   | 31:10 | Reserved                                                                |     ?    |
   +-------+-------------------------------------------------------------------------+----------+


U0LCR - UART0 Line Control Register (address 0xE000 C00C)
   +-------+-------------------------------------------------------------------------+----------+
   |  bit  |                                                                         |Reset val.|
   +-------+-------------------------------------------------------------------------+----------+
   |  1:0  | Word Length Select:                                                     |     0    |
   |       | 00 -> 5 bit character length         01 -> 6 bit character length       |          |
   |       | 10 -> 7 bit character length         11 -> 8 bit character length       |          |
   +-------+-------------------------------------------------------------------------+----------+
   |   2   | Stop Bit Select:                                                        |     0    |
   |       | 0 -> 1 stop bit               1 -> 2 stop bits (1.5 dla 5 bit)          |          |
   +-------+-------------------------------------------------------------------------+----------+
   |   3   | Parity Enable:                                                          |     0    |
   |       | 0 -> Disable parity generation and checking.                            |          |
   |       | 1 -> Enable parity generation and checking.                             |          |
   +-------+-------------------------------------------------------------------------+----------+
   |  5:4  | Parity Select:                                                          |     0    |
   |       | 00 -> Odd parity                     01 -> Even Parity                  |          |
   |       | 10 -> Forced "1" stick parity        11 -> Forced "0" stick parity      |          |
   +-------+-------------------------------------------------------------------------+----------+
   |   6   | Break Control:                                                          |     0    |
   |       |   0 -> Disable break transmission    1 -> Enable break transmission     |          |
   +-------+-------------------------------------------------------------------------+----------+
   |   7   | Divisor Latch Access Bit (DLAB)                                         |     0    |
   |       | 0 -> Disable access to Divisor Latches.                                 |          |
   |       | 1 -> Enable access to Divisor Latches.                                  |          |
   +-------+-------------------------------------------------------------------------+----------+


U0LSR - UART0 Line Status Register (address 0xE000 C014)
   +-------+-------------------------------------------------------------------------+----------+
   |  bit  |                                                                         |Reset val.|
   +-------+-------------------------------------------------------------------------+----------+
   |    0  | Receiver Data Ready (RDR)                                               |     0    |
   |       | U0LSR is set when the U0RBR holds an unread character and is cleared    |          |
   |       | when the UART0 RBR FIFO is empty.                                       |          |
   |       | 0 -> U0RBR is empty.                                                    |          |
   |       | 1 -> U0RBR contains valid data.                                         |          |
   +-------+-------------------------------------------------------------------------+----------+
   |    1  | Overrun Error (OE)                                                      |     0    |
   |       | The overrun error condition is set as soon as it occurs. An U0LSR read  |          |
   |       | clears U0LSR1. U0LSR1 is set when UART0 RSR has a new character         |          |
   |       | assembled and the UART0 RBR FIFO is full. In this case, the UART0 RBR   |          |
   |       | FIFO will not be overwritten and the character in the UART0 RSR will    |          |
   |       | be lost.                                                                |          |
   |       | 0 -> Overrun error status is inactive.                                  |          |
   |       | 1 -> Overrun error status is active.                                    |          |
   +-------+-------------------------------------------------------------------------+----------+
   |    2  | Parity Error (PE)                                                       |     0    |
   |       | When the parity bit of a received character is in the wrong state, a    |          |
   |       | parity error occurs. An U0LSR read clears U0LSR. Time of parity error   |          |
   |       | detection is dependent on U0FCR.                                        |          |
   |       | 0 -> Parity error status is inactive.                                   |          |
   |       | 1 -> Parity error status is active.                                     |          |
   +-------+-------------------------------------------------------------------------+----------+
   |    3  | Framing Error (FE)                                                      |     0    |
   |       | When the stop bit of a received character is a logic 0, a framing error |          |
   |       | occurs. An U0LSR read clears U0LSR.                                     |          |
   |       | 0 -> Framing error status is inactive.                                  |          |
   |       | 1 -> Framing error status is active.                                    |          |
   +-------+-------------------------------------------------------------------------+----------+
   |    4  | Break Interrupt (BI)                                                    |     0    |
   |       | When RXD0 is held in the spacing state (all 0s) for one full character |          |
   |       | transmission (start, data, parity, stop), a break interrupt occurs.     |          |
   |       | 0 -> Break interrupt status is inactive.                                |          |
   |       | 1 -> Break interrupt status is active.                                  |          |
   +-------+-------------------------------------------------------------------------+----------+
   |    5  | Transmitter Holding Register Empty (THRE)                               |     0    |
   |       | THRE is set immediately upon detection of an empty UART0 THR and is     |          |
   |       | cleared on a U0THR write.                                               |          |
   |       | 1 -> U0THR contains valid data.                                         |          |
   |       | 1 U0THR is empty.                                                       |          |
   +-------+-------------------------------------------------------------------------+----------+
   |    6  | Transmitter Empty (TEMT)                                                |     0    |
   |       | TEMT is set when both U0THR and U0TSR are empty; TEMT is cleared when   |          |
   |       | either the U0TSR or the U0THR contain valid data.                       |          |
   |       | 1 -> U0THR and/or the U0TSR contains valid data.                        |          |
   |       | 1 -> U0THR and the U0TSR are empty.                                     |          |
   +-------+-------------------------------------------------------------------------+----------+
   |    7  | Error in RX FIFO (RXFE)                                                 |     0    |
   |       | U0LSR is set when a character with a Rx error such as framing error,    |          |
   |       | parity error or break interrupt, is loaded into the U0RBR. This bit is  |          |
   |       | cleared when the U0LSR register is read and there are no subsequent     |          |
   |       | errors in the UART0 FIFO.                                               |          |
   |       | 0 -> U0RBR contains no UART0 RX errors or U0FCR=0.                      |          |
   |       | 1 -> UART0 RBR contains at least one UART0 RX error.                    |          |
   +-------+-------------------------------------------------------------------------+----------+



U0SCR - UART0 Scratch Pad Register (address 0xE000 C01C)
   +-------+-------------------------------------------------------------------------+----------+
   |  bit  |                                                                         |Reset val.|
   +-------+-------------------------------------------------------------------------+----------+
   |  7:0  | A readable, writable byte.                                              |    0x00  |
   +-------+-------------------------------------------------------------------------+----------+


U0ACR - UART0 Auto-baud Control Register (0xE000 C020)
   +-------+-------------------------------------------------------------------------+----------+
   |  bit  |                                                                         |Reset val.|
   +-------+-------------------------------------------------------------------------+----------+
   |   0   | Start This bit is automatically cleared after auto-baud completion.     |     0    |
   |       | 0 -> Auto-baud stop (auto-baud is not running).                         |          |
   |       | 1 -> Auto-baud start (auto-baud is running). Auto-baud run bit.         |          |
   |       | This bit is automatically cleared after auto-baud completion.           |          |
   +-------+-------------------------------------------------------------------------+----------+
   |   1   | Mode Auto-baud mode select bit.                                         |     0    |
   |       | 0 -> Mode 0.                                                            |          |
   |       | 1 -> Mode 1.                                                            |          |
   +-------+-------------------------------------------------------------------------+----------+
   |   2   | AutoRestart                                                             |     0    |
   |       | 0 -> No restart.                                                        |          |
   |       | 1 -> Restart in case of time-out (counter restarts at next UART0 Rx     |          |
   |       | falling edge)                                                           |          |
   +-------+-------------------------------------------------------------------------+----------+
   |  7:3  | Reserved                                                                |          |
   +-------+-------------------------------------------------------------------------+----------+
   |   8   | ABEOIntClr End of auto-baud interrupt clear bit (write only accessible).|     0    |
   |       | Writing a 1 will clear the corresponding interrupt in the U0IIR.        |          |
   |       | Writing a 0 has no impact.                                              |          |
   +-------+-------------------------------------------------------------------------+----------+
   |   9   | ABTOIntClr Auto-baud time-out interrupt clear bit (write only           |     0    |
   |       | accessible). Writing a 1 will clear the corresponding interrupt in the  |          |
   |       | U0IIR. Writing a 0 has no impact.                                       |          |
   +-------+-------------------------------------------------------------------------+----------+
   | 31:10 | Reserved                                                                |          |
   +-------+-------------------------------------------------------------------------+----------+
*/


static void TraceTxDSignal ( void )
{
  /*-------------------------------------------------------------------------*/
  UART0TxdLEDDelayCounter = DefaultLEDDelay ;
  LEDOn ( UART0TxDLED ) ;
} /* TraceTxDSignal */


static void TraceRxDSignal ( void )
{
  /*-------------------------------------------------------------------------*/
  UART0RxDLEDDelayCounter = DefaultLEDDelay ;
  LEDOn ( UART0RxDLED ) ;
} /* TraceRxDSignal */


__irq void UART0IrqHandler ( void )
{
  UCHAR Data ;
  UCHAR U0IrqStatus ;
  /*-------------------------------------------------------------------------*/
  for ( ; ; )
  {
    U0IrqStatus = ( UCHAR ) U0IIR ;
    if ( U0IrqStatus & 0x01 )
    {
      VICVectAddr = 0 ;
      return ;
    } /* if */ ;
    switch ( ( U0IrqStatus >> 1 ) & 0x03 )
    {
      case 1             :   /* 0 1 - transmitter */
        if ( UART0OutBuffer . RdDataPointer == UART0OutBuffer . WrDataPointer )
        {
          UART0OutBuffer . BusyStatus = 0 ;
          break ;
        } /* if */ ;
        Data = UART0OutBuffer . Buffer [ UART0OutBuffer . RdDataPointer ] ;
        UART0OutBuffer . RdDataPointer ++ ;
        if ( UART0OutBuffer . RdDataPointer >= OutSerialBuffSize )
          UART0OutBuffer . RdDataPointer = 0 ;
        U0THR = Data ;
        TraceTxDSignal ( ) ;
        break ;
      case 2             :   /* 1 0 - reciever data */
        Data = U0RBR ;
        UART0InpBuffer . Buffer [ UART0InpBuffer . WrDataPointer ] = Data ;
        UART0InpBuffer . WrDataPointer ++ ;
        if ( UART0InpBuffer . WrDataPointer >= InpSerialBuffSize )
          UART0InpBuffer . WrDataPointer = 0 ;
        TraceRxDSignal ( )  ;
        break ;
      default            :
        VICVectAddr = 0 ;
        return ;
    } /* switch */ ;
  } /* for */ ;
} /* UART0IrqHandler */


void UART0InputReset ( void )
{
  /*-------------------------------------------------------------------------*/
  UART0InpBuffer . RdDataPointer = 0 ;
  UART0InpBuffer . WrDataPointer = 0 ;
} /* UART0InputReset */


static void UART0OutputReset ( void )
{
  /*-------------------------------------------------------------------------*/
  UART0OutBuffer . RdDataPointer = 0 ;
  UART0OutBuffer . WrDataPointer = 0 ;
  UART0OutBuffer . BusyStatus = 0 ;
} /* UART0OutputReset */


static void UART0DisableInterrupt ( void )
{
  /*-------------------------------------------------------------------------*/
  VICIntEnClr = UART0InterruptMask ;
} /* UART0DisableInterrupt */


static void UART0EnableInterrupt ( void )
{
  /*-------------------------------------------------------------------------*/
  VICIntEnable = UART0InterruptMask ;
} /* UART0EnableInterrupt */


void UART0Send ( UCHAR Data )
{
  USHORT WStop ;
  /*-------------------------------------------------------------------------*/
  UART0DisableInterrupt ( ) ;
  if ( ! UART0OutBuffer . BusyStatus )
  {
    U0THR = Data ;
    UART0OutBuffer . BusyStatus = 1 ;
    UART0EnableInterrupt ( ) ;
    TraceTxDSignal ( ) ;
  } /* if ... */
  else
  {
    for ( ; ; )
    {
      WStop = UART0OutBuffer . WrDataPointer + 1 ;
      if ( WStop >= OutSerialBuffSize )
        WStop = 0 ;
      if ( WStop != UART0OutBuffer . RdDataPointer )
      {
        UART0OutBuffer . Buffer [ UART0OutBuffer . WrDataPointer ] = Data ;
        UART0OutBuffer . WrDataPointer = WStop ;
        UART0EnableInterrupt ( ) ;
        break ;
      } /* if ... */
      UART0EnableInterrupt ( ) ;
      NOP ( ) ;
      NOP ( ) ;
      NOP ( ) ;
      NOP ( ) ;
      NOP ( ) ;
      UART0DisableInterrupt ( ) ;
    } /* for */ ;
  } /* if ... else */ ;
} /* UART0Send */


void UART0SendString ( UCHAR * String )
{
  /*-------------------------------------------------------------------------*/
  while ( * String )
  {
    UART0Send ( * String ) ;
    String ++ ;
  } /* while */ ;
} /* UART0SendString */


ULONG UART0DataPresent ( void )
{
  ULONG Result ;
  /*-------------------------------------------------------------------------*/
  Result = FALSE ;
  UART0DisableInterrupt ( ) ;
  if ( UART0InpBuffer . RdDataPointer != UART0InpBuffer . WrDataPointer )
  {
    Result = TRUE ;
  } /* if */ ;
  UART0EnableInterrupt ( ) ;
  return ( Result ) ;
} /* UART0DataPresent */


UCHAR UART0GetData ( void )
{
  UCHAR Data ;
  /*-------------------------------------------------------------------------*/
  UART0DisableInterrupt ( ) ;
  if ( UART0InpBuffer . RdDataPointer == UART0InpBuffer . WrDataPointer )
  {
    UART0EnableInterrupt ( ) ;
    return ( 0 ) ;
  } /* if */ ;
  Data = UART0InpBuffer . Buffer [ UART0InpBuffer . RdDataPointer ] ;
  UART0InpBuffer . RdDataPointer ++ ;
  if ( UART0InpBuffer . RdDataPointer >= InpSerialBuffSize )
    UART0InpBuffer . RdDataPointer = 0 ;
  UART0EnableInterrupt ( ) ;
  return ( Data ) ;
} /* UARTGetData */


UCHAR UART0Working ( void )
{
  UCHAR Data ;
  /*-------------------------------------------------------------------------*/
  if ( UART0OutBuffer . BusyStatus )
    return FALSE ;
  Data = U0LSR ;
  if ( Data  & 0x40 )
    return ( TRUE ) ;
  else
    return ( FALSE ) ;
  } /* UART0Working */


void UART0HardwInit( ULONG PeriphClk ,
                     UCHAR SerialMode ,
                     ULONG SerialSpeed )
{
  ULONG Fdiv ;
  ULONG Fmod ;
  ULONG LDivider ;
  ULONG HDivider ;
  /*-------------------------------------------------------------------------*/
  UART0DisableInterrupt ( ) ;
  UART0InputReset ( ) ;
  UART0OutputReset ( ) ;
  Fdiv = ( PeriphClk / 16 ) / SerialSpeed ;
  Fmod = ( PeriphClk / 16 ) % SerialSpeed ;
  if ( ( Fmod << 1 ) >= SerialSpeed )
    Fdiv ++ ;
  LDivider = Fdiv % 256 ;
  HDivider = Fdiv / 256 ;
  U0FDR = 0x10 ;
  U0LCR = 0x80 ;
  U0DLL = LDivider ;
  U0DLM = HDivider ;
  U0LCR = SerialMode ;
  U0FCR = 0x07 ;
  VICVectAddr6  = ( unsigned long ) UART0IrqHandler ;
  VICVectCntl6  = 15 ;
  UART0EnableInterrupt ( ) ;
  U0IER = RecvDataInterruptEnable | TransmDataInterruptEnable ;
} /* UART0HardwInit */


void UART0InitEnvir ( void )
{
  /*-------------------------------------------------------------------------*/
  PINSEL0 &= ~ 0x000000F0 ;
  PINSEL0 |= 0x00000050 ;
  UART0InputReset ( ) ;
  UART0OutputReset ( ) ;
  UART0TxdLEDDelayCounter = 0 ;
  UART0RxDLEDDelayCounter = 0 ;
} /* UART0InitEnvir */


