/*
****************************************************************************
****************************************************************************
**                                                                        **
**                XXXX     X    X         X XXXXXX X                      **
**               X    X   X X   X    X    X X      X   X                  **
**               X       X   X  X   X X   X X      X  X                   **
**               X   XX XXXXXXX X  X   X  X XXXX   X X                    **
**               X    X X     X X X     X X X      XX                     **
**               X    X X     X XX       XX X      X                      **
**                XXXX  X     X X         X XXXXXX XXXXXX                 **
**                                                                        **
**                           BIALYSTOK 2009                               **
**                                                                        **
**                       Module: UART1 (interrupt)                        **
**                      Wersja z dnia : 2009.09.12                        **
**                     Napisal : Andrzej Pawluczuk                        **
**                                                                        **
****************************************************************************
****************************************************************************
*/

#include <LPC23xx.h>
#include "uart1.h"
#include "types.h"
#include "system.h"
#include "irqvectors.h"
#include "serialconst.h"
#include "leddisplay.h"

#define DefaultLEDDelay       44

#define UART1InterruptMask ( 1 << UART1_Interrupt )

#define InpSerialBuffSize     64
#define OutSerialBuffSize     64


typedef struct {
                   USHORT     RdDataPointer ;
                   USHORT     WrDataPointer ;
                   UCHAR      Buffer [ InpSerialBuffSize ] ;
                 } InpSerialBufferRecT ;

typedef struct {
                   USHORT     RdDataPointer ;
                   USHORT     WrDataPointer ;
                   USHORT     BusyStatus ;
                   UCHAR      Buffer [ OutSerialBuffSize ] ;
                 } OutSerialBufferRecT ;


static volatile InpSerialBufferRecT UART1InpBuffer ;
static volatile OutSerialBufferRecT UART1OutBuffer ;
UCHAR UART1TxdLEDDelayCounter ;
UCHAR UART1RxDLEDDelayCounter ;


static void TraceTxDSignal ( void )
{
  /*-------------------------------------------------------------------------*/
  UART1TxdLEDDelayCounter = DefaultLEDDelay ;
  LEDOn ( UART1TxDLED ) ;
} /* TraceTxDSignal */


static void TraceRxDSignal ( void )
{
  /*-------------------------------------------------------------------------*/
  UART1RxDLEDDelayCounter = DefaultLEDDelay ;
  LEDOn ( UART1RxDLED ) ;
} /* TraceRxDSignal */


__irq void UART1IrqHandler ( void )
{
  UCHAR Data ;
  UCHAR U1IrqStatus ;
  /*-------------------------------------------------------------------------*/
  for ( ; ; )
  {
    U1IrqStatus = ( UCHAR ) U1IIR ;
    if ( U1IrqStatus & 0x01 )
    {
      VICVectAddr = 0 ;
      return ;
    } /* if */ ;
    switch ( ( U1IrqStatus >> 1 ) & 0x03 )
    {
      case 1             :   /* 0 1 - transmitter */
        if ( UART1OutBuffer . RdDataPointer == UART1OutBuffer . WrDataPointer )
        {
          UART1OutBuffer . BusyStatus = 0 ;
          break ;
        } /* if */ ;
        Data = UART1OutBuffer . Buffer [ UART1OutBuffer . RdDataPointer ] ;
        UART1OutBuffer . RdDataPointer ++ ;
        if ( UART1OutBuffer . RdDataPointer >= OutSerialBuffSize )
          UART1OutBuffer . RdDataPointer = 0 ;
        U1THR = Data ;
        TraceTxDSignal ( ) ;
        break ;
      case 2             :   /* 1 0 - reciever data */
        Data = U1RBR ;
        UART1InpBuffer . Buffer [ UART1InpBuffer . WrDataPointer ] = Data ;
        UART1InpBuffer . WrDataPointer ++ ;
        if ( UART1InpBuffer . WrDataPointer >= InpSerialBuffSize )
          UART1InpBuffer . WrDataPointer = 0 ;
        TraceRxDSignal ( )  ;
        break ;
      default            :
        VICVectAddr = 0 ;
        return ;
    } /* switch */ ;
  } /* for */ ;
} /* UART1IrqHandler */


void UART1InputReset ( void )
{
  /*-------------------------------------------------------------------------*/
  UART1InpBuffer . RdDataPointer = 0 ;
  UART1InpBuffer . WrDataPointer = 0 ;
} /* UART1InputReset */


static void UART1OutputReset ( void )
{
  /*-------------------------------------------------------------------------*/
  UART1OutBuffer . RdDataPointer = 0 ;
  UART1OutBuffer . WrDataPointer = 0 ;
  UART1OutBuffer . BusyStatus = 0 ;
} /* UART1OutputReset */


static void UART1DisableInterrupt ( void )
{
  /*-------------------------------------------------------------------------*/
  VICIntEnClr = UART1InterruptMask ;
} /* UART1DisableInterrupt */


static void UART1EnableInterrupt ( void )
{
  /*-------------------------------------------------------------------------*/
  VICIntEnable = UART1InterruptMask ;
} /* UART1EnableInterrupt */


void UART1Send ( UCHAR Data )
{
  USHORT WStop ;
  /*-------------------------------------------------------------------------*/
  UART1DisableInterrupt ( ) ;
  if ( ! UART1OutBuffer . BusyStatus )
  {
    U1THR = Data ;
    UART1OutBuffer . BusyStatus = 1 ;
    UART1EnableInterrupt ( ) ;
    TraceTxDSignal ( ) ;
  } /* if ... */
  else
  {
    for ( ; ; )
    {
      WStop = UART1OutBuffer . WrDataPointer + 1 ;
      if ( WStop >= OutSerialBuffSize )
        WStop = 0 ;
      if ( WStop != UART1OutBuffer . RdDataPointer )
      {
        UART1OutBuffer . Buffer [ UART1OutBuffer . WrDataPointer ] = Data ;
        UART1OutBuffer . WrDataPointer = WStop ;
        UART1EnableInterrupt ( ) ;
        break ;
      } /* if ... */
      UART1EnableInterrupt ( ) ;
      NOP ( ) ;
      NOP ( ) ;
      NOP ( ) ;
      NOP ( ) ;
      NOP ( ) ;
      UART1DisableInterrupt ( ) ;
    } /* for */ ;
  } /* if ... else */ ;
} /* UART1Send */


void UART1SendString ( UCHAR * String )
{
  /*-------------------------------------------------------------------------*/
  while ( * String )
  {
    UART1Send ( * String ) ;
    String ++ ;
  } /* while */ ;
} /* UART1SendString */


ULONG UART1DataPresent ( void )
{
  ULONG Result ;
  /*-------------------------------------------------------------------------*/
  Result = FALSE ;
  UART1DisableInterrupt ( ) ;
  if ( UART1InpBuffer . RdDataPointer != UART1InpBuffer . WrDataPointer )
  {
    Result = TRUE ;
  } /* if */ ;
  UART1EnableInterrupt ( ) ;
  return ( Result ) ;
} /* UART1DataPresent */


UCHAR UART1GetData ( void )
{
  UCHAR Data ;
  /*-------------------------------------------------------------------------*/
  UART1DisableInterrupt ( ) ;
  if ( UART1InpBuffer . RdDataPointer == UART1InpBuffer . WrDataPointer )
  {
    UART1EnableInterrupt ( ) ;
    return ( 0 ) ;
  } /* if */ ;
  Data = UART1InpBuffer . Buffer [ UART1InpBuffer . RdDataPointer ] ;
  UART1InpBuffer . RdDataPointer ++ ;
  if ( UART1InpBuffer . RdDataPointer >= InpSerialBuffSize )
    UART1InpBuffer . RdDataPointer = 0 ;
  UART1EnableInterrupt ( ) ;
  return ( Data ) ;
} /* UARTGetData */


UCHAR UART1Working ( void )
{
  UCHAR Data ;
  /*-------------------------------------------------------------------------*/
  if ( UART1OutBuffer . BusyStatus )
    return FALSE ;
  Data = U1LSR ;
  if ( Data  & 0x40 )
    return ( TRUE ) ;
  else
    return ( FALSE ) ;
  } /* UART1Working */


void UART1HardwInit( ULONG PeriphClk ,
                     UCHAR SerialMode ,
                     ULONG SerialSpeed )
{
  ULONG Fdiv ;
  ULONG Fmod ;
  ULONG LDivider ;
  ULONG HDivider ;
  /*-------------------------------------------------------------------------*/
  UART1DisableInterrupt ( ) ;
  UART1InputReset ( ) ;
  UART1OutputReset ( ) ;
  Fdiv = ( PeriphClk / 16 ) / SerialSpeed ;
  Fmod = ( PeriphClk / 16 ) % SerialSpeed ;
  if ( ( Fmod << 1 ) >= SerialSpeed )
    Fdiv ++ ;
  LDivider = Fdiv % 256 ;
  HDivider = Fdiv / 256 ;
  U1FDR = 0x10 ;
  U1LCR = 0x80 ;
  U1DLL = LDivider ;
  U1DLM = HDivider ;
  U1LCR = SerialMode ;
  U1FCR = 0x07 ;
  VICVectAddr7  = ( unsigned long ) UART1IrqHandler ;
  VICVectCntl7  = 15 ;
  UART1EnableInterrupt ( ) ;
  U1IER = RecvDataInterruptEnable | TransmDataInterruptEnable ;
} /* UART1HardwInit */


void UART1InitEnvir ( void )
{
  /*-------------------------------------------------------------------------*/
  PINSEL4 &= ~ 0x0000FFFF ;         /* TXD1, RXD1, CTS1, DCD1, DSR1, DTR1, RI1, RTS1 */
  PINSEL4 |= 0x0000AAAA ;
  UART1InputReset ( ) ;
  UART1OutputReset ( ) ;
  UART1TxdLEDDelayCounter = 0 ;
  UART1RxDLEDDelayCounter = 0 ;
} /* UART1InitEnvir */
